# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

import os

import numpy as np
import pytest
import qiskit
from cirq import (
    Circuit,
    CXPowGate,
    DepolarizingChannel,
    I,
    LineQubit,
    MixedUnitaryChannel,
    Rx,
    Rz,
    X,
    Y,
    Z,
    ops,
    unitary,
)

from mitiq import Executor, Observable, PauliString
from mitiq.cdr import generate_training_circuits
from mitiq.cdr._testing import random_x_z_cnot_circuit
from mitiq.interface.mitiq_cirq import compute_density_matrix
from mitiq.interface.mitiq_qiskit import qiskit_utils
from mitiq.interface.mitiq_qiskit.conversions import to_qiskit
from mitiq.pec.representations.learning import (
    _parse_learning_kwargs,
    biased_noise_loss_function,
    depolarizing_noise_loss_function,
    learn_biased_noise_parameters,
    learn_depolarizing_noise_parameter,
)

rng = np.random.RandomState(1)
circuit = random_x_z_cnot_circuit(
    LineQubit.range(2), n_moments=5, random_state=rng
)

# Set number of samples used to calculate mitigated value in loss function
pec_kwargs = {"num_samples": 20, "random_state": 1}

observable = Observable(PauliString("XZ"), PauliString("YY"))

training_circuits = generate_training_circuits(
    circuit=circuit,
    num_training_circuits=3,
    fraction_non_clifford=0,
    method_select="uniform",
    method_replace="closest",
)

CNOT_ops = list(circuit.findall_operations_with_gate_type(CXPowGate))
Rx_ops = list(circuit.findall_operations_with_gate_type(Rx))
Rz_ops = list(circuit.findall_operations_with_gate_type(Rz))


def ideal_execute(circ: Circuit) -> np.ndarray:
    return compute_density_matrix(circ, noise_level=(0.0,))


ideal_executor = Executor(ideal_execute)
ideal_values = np.array(ideal_executor.evaluate(training_circuits, observable))


def biased_noise_channel(epsilon: float, eta: float) -> MixedUnitaryChannel:
    a = 1 - epsilon
    b = epsilon * (3 * eta + 1) / (3 * (eta + 1))
    c = epsilon / (3 * (eta + 1))

    mix = [
        (a, unitary(I)),
        (b, unitary(Z)),
        (c, unitary(X)),
        (c, unitary(Y)),
    ]
    return ops.MixedUnitaryChannel(mix)


@pytest.mark.parametrize("epsilon", [0.05, 0.1])
@pytest.mark.parametrize(
    "operations", [[Circuit(CNOT_ops[0][1])], [Circuit(Rx_ops[0][1])]]
)
def test_depolarizing_noise_loss_function(epsilon, operations):
    """Test that the biased noise loss function value (calculated with error
    mitigation) is less than (or equal to) the loss calculated with the noisy
    (unmitigated) executor"""

    def noisy_execute(circ: Circuit) -> np.ndarray:
        noisy_circ = circ.with_noise(DepolarizingChannel(epsilon))
        return ideal_execute(noisy_circ)

    noisy_executor = Executor(noisy_execute)
    noisy_values = np.array(
        noisy_executor.evaluate(training_circuits, observable)
    )
    loss = depolarizing_noise_loss_function(
        epsilon=np.array([epsilon]),
        operations_to_mitigate=operations,
        training_circuits=training_circuits,
        ideal_values=ideal_values,
        noisy_executor=noisy_executor,
        pec_kwargs=pec_kwargs,
        observable=observable,
    )

    assert loss <= np.mean((noisy_values - ideal_values) ** 2)


@pytest.mark.parametrize("epsilon", [0, 0.7, 1])
@pytest.mark.parametrize("eta", [0, 1])
@pytest.mark.parametrize(
    "operations", [[Circuit(CNOT_ops[0][1])], [Circuit(Rx_ops[0][1])]]
)
def test_biased_noise_loss_function(epsilon, eta, operations):
    """Test that the biased noise loss function value (calculated with error
    mitigation) is less than (or equal to) the loss calculated with the noisy
    (unmitigated) executor"""

    def noisy_execute(circ: Circuit) -> np.ndarray:
        noisy_circ = circ.with_noise(biased_noise_channel(epsilon, eta))
        return ideal_execute(noisy_circ)

    noisy_executor = Executor(noisy_execute)
    noisy_values = np.array(
        noisy_executor.evaluate(training_circuits, observable)
    )

    loss = biased_noise_loss_function(
        params=[epsilon, eta],
        operations_to_mitigate=operations,
        training_circuits=training_circuits,
        ideal_values=ideal_values,
        noisy_executor=noisy_executor,
        pec_kwargs=pec_kwargs,
        observable=observable,
    )

    assert loss <= np.mean((noisy_values - ideal_values) ** 2)


@pytest.mark.parametrize(
    "operations", [[Circuit(CNOT_ops[0][1])], [Circuit(Rz_ops[0][1])]]
)
def test_biased_noise_loss_compare_ideal(operations):
    """Test that the loss function is zero when the noise strength is zero"""

    def noisy_execute(circ: Circuit) -> np.ndarray:
        noisy_circ = circ.with_noise(biased_noise_channel(0, 0))
        return ideal_execute(noisy_circ)

    noisy_executor = Executor(noisy_execute)

    loss = biased_noise_loss_function(
        params=[0, 0],
        operations_to_mitigate=operations,
        training_circuits=training_circuits,
        ideal_values=ideal_values,
        noisy_executor=noisy_executor,
        pec_kwargs=pec_kwargs,
        observable=observable,
    )
    assert np.isclose(loss, 0)


@pytest.mark.parametrize(
    "operations",
    [
        [to_qiskit(Circuit(CNOT_ops[0][1]))],
        [to_qiskit(Circuit(Rx_ops[0][1]))],
        [to_qiskit(Circuit(Rz_ops[0][1]))],
    ],
)
def test_biased_noise_loss_function_qiskit(operations):
    """Test the learning function with initial noise strength and noise bias
    with a small offset from the simulated noise model values"""
    qiskit_circuit = to_qiskit(circuit)

    qiskit_training_circuits = generate_training_circuits(
        circuit=qiskit_circuit,
        num_training_circuits=3,
        fraction_non_clifford=0.2,
        method_select="uniform",
        method_replace="closest",
        random_state=rng,
    )
    obs = Observable(PauliString("XY"), PauliString("ZZ"))

    def ideal_execute_qiskit(circ: qiskit.QuantumCircuit) -> float:
        return qiskit_utils.execute(circ, obs.matrix())

    ideal_executor_qiskit = Executor(ideal_execute_qiskit)
    ideal_values = np.array(
        ideal_executor_qiskit.evaluate(qiskit_training_circuits)
    )

    epsilon = 0.1

    def noisy_execute_qiskit(circ: qiskit.QuantumCircuit) -> float:
        noise_model = qiskit_utils.initialized_depolarizing_noise(epsilon)
        return qiskit_utils.execute_with_noise(circ, obs.matrix(), noise_model)

    noisy_executor_qiskit = Executor(noisy_execute_qiskit)

    noisy_values = np.array(
        noisy_executor_qiskit.evaluate(qiskit_training_circuits)
    )

    loss = biased_noise_loss_function(
        params=[epsilon, 0],
        operations_to_mitigate=operations,
        training_circuits=qiskit_training_circuits,
        ideal_values=ideal_values,
        noisy_executor=noisy_executor_qiskit,
        pec_kwargs=pec_kwargs,
    )

    assert loss <= np.mean((noisy_values - ideal_values) ** 2)


@pytest.mark.parametrize("epsilon", [0.05, 0.1])
def test_learn_depolarizing_noise_parameter(epsilon):
    """Test the learning function with initial noise strength with a small
    offset from the simulated noise model values"""

    operations_to_learn = [Circuit(op[1]) for op in CNOT_ops]

    offset = 0.1

    def noisy_execute(circ: Circuit) -> np.ndarray:
        noisy_circ = circ.copy()
        insertions = []
        for op in CNOT_ops:
            index = op[0] + 1
            qubits = op[1].qubits
            for q in qubits:
                insertions.append((index, DepolarizingChannel(epsilon)(q)))
        noisy_circ.batch_insert(insertions)
        return ideal_execute(noisy_circ)

    noisy_executor = Executor(noisy_execute)

    epsilon0 = (1 - offset) * epsilon
    eps_string = str(epsilon).replace(".", "_")
    pec_data = np.loadtxt(
        os.path.join(
            "./mitiq/pec/representations/tests/learning_pec_data",
            f"learning_pec_data_eps_{eps_string}.txt",
        )
    )

    [success, epsilon_opt] = learn_depolarizing_noise_parameter(
        operations_to_learn=operations_to_learn,
        circuit=circuit,
        ideal_executor=ideal_executor,
        noisy_executor=noisy_executor,
        num_training_circuits=5,
        fraction_non_clifford=0.2,
        training_random_state=np.random.RandomState(1),
        epsilon0=epsilon0,
        observable=observable,
        learning_kwargs={"pec_data": pec_data},
    )
    assert success
    assert abs(epsilon_opt - epsilon) < offset * epsilon


@pytest.mark.parametrize("epsilon", [0.05, 0.1])
@pytest.mark.parametrize("eta", [1, 2])
def test_learn_biased_noise_parameters(epsilon, eta):
    """Test the learning function can run without pre-executed data"""

    operations_to_learn = [Circuit(op[1]) for op in CNOT_ops]

    def noisy_execute(circ: Circuit) -> np.ndarray:
        noisy_circ = circ.copy()
        insertions = []
        for op in CNOT_ops:
            index = op[0] + 1
            qubits = op[1].qubits
            for q in qubits:
                insertions.append(
                    (index, biased_noise_channel(epsilon, eta)(q))
                )
        noisy_circ.batch_insert(insertions)
        return ideal_execute(noisy_circ)

    noisy_executor = Executor(noisy_execute)

    eps_offset = 0.1
    eta_offset = 0.2
    epsilon0 = (1 - eps_offset) * epsilon
    eta0 = (1 - eta_offset) * eta

    num_training_circuits = 5
    pec_data = np.zeros([122, 122, num_training_circuits])

    eps_string = str(epsilon).replace(".", "_")
    for tc in range(0, num_training_circuits):
        pec_data[:, :, tc] = np.loadtxt(
            os.path.join(
                "./mitiq/pec/representations/tests/learning_pec_data",
                f"learning_pec_data_eps_{eps_string}eta_{eta}tc_{tc}.txt",
            )
        )

    [success, epsilon_opt, eta_opt] = learn_biased_noise_parameters(
        operations_to_learn=operations_to_learn,
        circuit=circuit,
        ideal_executor=ideal_executor,
        noisy_executor=noisy_executor,
        num_training_circuits=num_training_circuits,
        fraction_non_clifford=0.2,
        training_random_state=np.random.RandomState(1),
        epsilon0=epsilon0,
        eta0=eta0,
        observable=observable,
        learning_kwargs={"pec_data": pec_data},
    )
    assert success
    assert abs(epsilon_opt - epsilon) < eps_offset * epsilon
    assert abs(eta_opt - eta) < eta_offset * eta


def test_empty_learning_kwargs():
    learning_kwargs = {}
    pec_data, method, minimize_kwargs = _parse_learning_kwargs(
        learning_kwargs=learning_kwargs
    )
    assert pec_data is None
    assert method == "Nelder-Mead"
    assert minimize_kwargs == {}



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Unit tests for quantum processing functions for classical shadows."""

import importlib
from typing import Callable, List
from unittest.mock import patch

import cirq
import cirq.testing
import pytest
from qiskit_aer import Aer

import mitiq
from mitiq import MeasurementResult
from mitiq.interface.mitiq_cirq.cirq_utils import (
    sample_bitstrings as cirq_sample_bitstrings,
)
from mitiq.interface.mitiq_qiskit.conversions import to_qiskit
from mitiq.interface.mitiq_qiskit.qiskit_utils import (
    sample_bitstrings as qiskit_sample_bitstrings,
)
from mitiq.shadows.quantum_processing import (
    generate_random_pauli_strings,
    get_rotated_circuits,
    random_pauli_measurement,
)


def test_tqdm_import_available():
    # Test the case where tqdm is available
    import tqdm as tqdm_orig

    assert tqdm_orig is not None
    assert mitiq.shadows.quantum_processing.tqdm is not None


def test_tqdm_import_not_available():
    with patch.dict("sys.modules", {"tqdm": None}):
        importlib.reload(
            mitiq.shadows.quantum_processing
        )  # Reload the module to trigger the import
        assert mitiq.shadows.quantum_processing.tqdm is None

    # Reload the module again to restore the original tqdm import.
    # Otherwise, the rest of the tests are affected by the patch (issue #2318)
    importlib.reload(mitiq.shadows.quantum_processing)


def test_generate_random_pauli_strings():
    """Tests that the function generates random Pauli strings."""
    num_qubits = 5
    num_strings = 10

    # Generate random pauli strings
    result = generate_random_pauli_strings(num_qubits, num_strings)

    # Check that the result is a list
    assert isinstance(result, List)

    # Check that the number of strings matches the input
    assert len(result) == num_strings

    # Check that each string has the right number of qubits
    for pauli_string in result:
        assert len(pauli_string) == num_qubits

    # Check that each string contains only the letters X, Y, and Z
    for pauli_string in result:
        assert set(pauli_string).issubset(set(["X", "Y", "Z"]))

    # Check that the function raises an exception for negative num_qubits
    # or num_strings
    with pytest.raises(ValueError):
        generate_random_pauli_strings(-1, num_strings)

    with pytest.raises(ValueError):
        generate_random_pauli_strings(num_qubits, -1)


def cirq_executor(circuit: cirq.Circuit) -> MeasurementResult:
    return cirq_sample_bitstrings(
        circuit,
        noise_level=(0,),
        shots=1,
        sampler=cirq.Simulator(),
    )


def qiskit_executor(circuit: cirq.Circuit) -> MeasurementResult:
    return qiskit_sample_bitstrings(
        to_qiskit(circuit),
        noise_model=None,
        backend=Aer.get_backend("aer_simulator"),
        shots=1,
        measure_all=False,
    )


def test_get_rotated_circuits():
    """Tests that the circuit is rotated."""

    # define circuit
    circuit = cirq.Circuit()
    qubits = cirq.LineQubit.range(4)
    circuit.append(cirq.H(qubits[0]))
    circuit.append(cirq.CNOT(qubits[0], qubits[1]))
    circuit.append(cirq.CNOT(qubits[1], qubits[2]))
    circuit.append(cirq.CNOT(qubits[2], qubits[3]))

    # define the pauli measurements to be performed on the circuit
    pauli_strings = ["XYZX"]
    # Rotate the circuit.
    rotated_circuits = get_rotated_circuits(circuit, pauli_strings)
    # Verify that the circuit was rotated.
    circuit_1 = circuit.copy()
    circuit_1.append(cirq.H(qubits[0]))
    circuit_1.append(cirq.S(qubits[1]) ** -1)
    circuit_1.append(cirq.H(qubits[1]))
    circuit_1.append(cirq.H(qubits[3]))
    circuit_1.append(cirq.measure(*qubits))

    assert rotated_circuits[0] == circuit_1

    for rc in rotated_circuits:
        assert isinstance(rc, cirq.Circuit)


# define a simple test circuit for the following tests
def simple_test_circuit(qubits):
    circuit = cirq.Circuit()
    num_qubits = len(qubits)
    circuit.append(cirq.H.on_each(*qubits))
    for i in range(num_qubits - 1):
        circuit.append(cirq.CNOT(qubits[i], qubits[i + 1]))
    return circuit


@pytest.mark.parametrize("n_qubits", [1, 2, 5])
@pytest.mark.parametrize("executor", [cirq_executor, qiskit_executor])
def test_random_pauli_measurement_no_errors(n_qubits, executor):
    """Test that random_pauli_measurement runs without errors."""
    qubits = cirq.LineQubit.range(n_qubits)
    circuit = simple_test_circuit(qubits)
    random_pauli_measurement(
        circuit, n_total_measurements=10, executor=executor
    )


@pytest.mark.parametrize("n_qubits", [1, 2, 5])
@pytest.mark.parametrize("executor", [cirq_executor, qiskit_executor])
def test_random_pauli_measurement_output_dimensions(
    n_qubits: int, executor: Callable
):
    """Test that random_pauli_measurement returns the correct output
    dimensions."""
    qubits = cirq.LineQubit.range(n_qubits)
    circuit = simple_test_circuit(qubits)
    n_total_measurements = 10
    shadow_outcomes, pauli_strings = random_pauli_measurement(
        circuit, n_total_measurements, executor=executor
    )
    shadow_outcomes_shape = len(shadow_outcomes), len(shadow_outcomes[0])
    pauli_strings_shape = len(pauli_strings), len(pauli_strings[0])
    assert shadow_outcomes_shape == (n_total_measurements, n_qubits), (
        f"Shadow outcomes have incorrect shape, expected "
        f"{(n_total_measurements, n_qubits)}, got {shadow_outcomes_shape}"
    )
    assert pauli_strings_shape == (n_total_measurements, n_qubits), (
        f"Pauli strings have incorrect shape, expected "
        f"{(n_total_measurements, n_qubits)}, got {pauli_strings_shape}"
    )


@pytest.mark.parametrize("n_qubits", [1, 2, 5])
@pytest.mark.parametrize("executor", [cirq_executor, qiskit_executor])
def test_random_pauli_measurement_output_types(
    n_qubits: int, executor: Callable
):
    """Test that random_pauli_measurement returns the correct output types."""
    qubits = cirq.LineQubit.range(n_qubits)
    circuit = simple_test_circuit(qubits)
    shadow_outcomes, pauli_strings = random_pauli_measurement(
        circuit, n_total_measurements=10, executor=executor
    )
    assert isinstance(shadow_outcomes[0], str)
    assert isinstance(pauli_strings[0], str)



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Unit tests for zero-noise extrapolation."""

import functools
import random
from typing import List
from unittest.mock import Mock

import cirq
import numpy as np
import pytest
import qiskit
import qiskit.circuit
from qiskit_aer import AerSimulator

from mitiq import QPROGRAM, SUPPORTED_PROGRAM_TYPES
from mitiq.benchmarks.randomized_benchmarking import generate_rb_circuits
from mitiq.interface import accept_any_qprogram_as_input, convert_from_mitiq
from mitiq.interface.mitiq_braket import to_braket
from mitiq.interface.mitiq_cirq import (
    compute_density_matrix,
    sample_bitstrings,
)
from mitiq.interface.mitiq_pennylane import to_pennylane
from mitiq.interface.mitiq_pyquil import to_pyquil
from mitiq.interface.mitiq_qibo import to_qibo
from mitiq.interface.mitiq_qiskit import (
    execute_with_shots_and_noise,
    initialized_depolarizing_noise,
    to_qiskit,
)
from mitiq.observable import Observable, PauliString
from mitiq.zne import (
    execute_with_zne,
    inference,
    mitigate_executor,
    scaling,
    zne_decorator,
)
from mitiq.zne.inference import (
    AdaExpFactory,
    LinearFactory,
    PolyFactory,
    RichardsonFactory,
)
from mitiq.zne.scaling import (
    fold_all,
    fold_gates_at_random,
    fold_global,
    get_layer_folding,
    insert_id_layers,
)
from mitiq.zne.zne import combine_results, construct_circuits

BASE_NOISE = 0.007
TEST_DEPTH = 30
ONE_QUBIT_GS_PROJECTOR = np.array([[1, 0], [0, 0]])

npX = np.array([[0, 1], [1, 0]])
"""Defines the sigma_x Pauli matrix in SU(2) algebra as a (2,2) `np.array`."""

npZ = np.array([[1, 0], [0, -1]])
"""Defines the sigma_z Pauli matrix in SU(2) algebra as a (2,2) `np.array`."""

# Default qubit register and circuit for unit tests
qreg = cirq.GridQubit.rect(2, 1)
circ = cirq.Circuit(cirq.ops.H.on_each(*qreg), cirq.measure_each(*qreg))


@accept_any_qprogram_as_input
def generic_executor(circuit, noise_level: float = 0.1) -> float:
    """Executor that simulates a circuit of any type and returns
    the expectation value of the ground state projector."""
    noisy_circuit = circuit.with_noise(cirq.depolarize(p=noise_level))
    result = cirq.DensityMatrixSimulator().simulate(noisy_circuit)
    return result.final_density_matrix[0, 0].real


# Default executor for unit tests
def executor(circuit) -> float:
    wavefunction = circuit.final_state_vector(
        ignore_terminal_measurements=True
    )
    return np.real(wavefunction.conj().T @ np.kron(npX, npZ) @ wavefunction)


@pytest.mark.parametrize(
    "executor", (sample_bitstrings, compute_density_matrix)
)
def test_with_observable_batched_factory(executor):
    observable = Observable(PauliString(spec="Z"))
    circuit = cirq.Circuit(cirq.H.on(cirq.LineQubit(0))) * 20
    executor = functools.partial(
        executor, noise_model_function=cirq.depolarize
    )

    real_factory = PolyFactory(scale_factors=[1, 3, 5], order=2)
    mock_factory = Mock(spec_set=PolyFactory, wraps=real_factory)
    zne_val = execute_with_zne(
        circuit,
        executor=executor,
        observable=observable,
        factory=mock_factory,
    )

    mock_factory.run.assert_called_with(
        circuit, executor, observable, fold_gates_at_random, 1
    )
    assert 0 <= zne_val <= 2


@pytest.mark.parametrize(
    "executor", (sample_bitstrings, compute_density_matrix)
)
def test_with_observable_adaptive_factory(executor):
    observable = Observable(PauliString(spec="Z"))
    circuit = cirq.Circuit(cirq.H.on(cirq.LineQubit(0))) * 20

    noisy_value = observable.expectation(circuit, sample_bitstrings)
    zne_value = execute_with_zne(
        circuit,
        executor=functools.partial(
            executor, noise_model_function=cirq.amplitude_damp
        ),
        observable=observable,
        factory=AdaExpFactory(steps=4, asymptote=0.5),
    )
    true_value = observable.expectation(
        circuit, functools.partial(compute_density_matrix, noise_level=(0,))
    )

    assert abs(zne_value - true_value) <= abs(noisy_value - true_value)


def test_with_observable_two_qubits():
    observable = Observable(
        PauliString(spec="XX", coeff=-1.21), PauliString(spec="ZZ", coeff=0.7)
    )
    circuit = cirq.Circuit(
        cirq.H.on(cirq.LineQubit(0)), cirq.CNOT.on(*cirq.LineQubit.range(2))
    )
    circuit += [circuit.copy(), cirq.inverse(circuit.copy())] * 20
    executor = compute_density_matrix

    noisy_value = observable.expectation(circuit, executor)
    zne_value = execute_with_zne(
        circuit,
        executor=functools.partial(
            executor, noise_model_function=cirq.depolarize
        ),
        observable=observable,
        factory=PolyFactory(scale_factors=[1, 3, 5], order=2),
    )
    true_value = observable.expectation(
        circuit, functools.partial(executor, noise_level=(0,))
    )

    assert abs(zne_value - true_value) <= 3 * abs(noisy_value - true_value)


@pytest.mark.parametrize(
    "fold_method",
    [
        fold_gates_at_random,
        insert_id_layers,
    ],
)
@pytest.mark.parametrize("factory", [LinearFactory, RichardsonFactory])
@pytest.mark.parametrize("num_to_average", [1, 2, 5])
def test_execute_with_zne_no_noise(fold_method, factory, num_to_average):
    """Tests execute_with_zne with noiseless simulation."""
    zne_value = execute_with_zne(
        circ,
        executor,
        num_to_average=num_to_average,
        scale_noise=fold_method,
        factory=factory([1.0, 2.0, 3.0]),
    )
    assert np.isclose(zne_value, 0.0)


@pytest.mark.parametrize("factory", [LinearFactory, RichardsonFactory])
@pytest.mark.parametrize(
    "fold_method",
    [
        fold_gates_at_random,
        insert_id_layers,
    ],
)
def test_averaging_improves_zne_value_with_fake_noise(factory, fold_method):
    """Tests that averaging with Gaussian noise produces a better ZNE value
    compared to not averaging with several folding methods.

    For non-deterministic folding, the ZNE value with average should be better.
    For deterministic folding, the ZNE value should be the same.
    """
    for seed in range(5):
        rng = np.random.RandomState(seed)

        def noisy_executor(circuit) -> float:
            return executor(circuit) + rng.randn()

        zne_value_no_averaging = execute_with_zne(
            circ,
            noisy_executor,
            num_to_average=1,
            scale_noise=fold_gates_at_random,
            factory=factory([1.0, 2.0, 3.0]),
        )

        zne_value_averaging = execute_with_zne(
            circ,
            noisy_executor,
            num_to_average=10,
            scale_noise=fold_method,
            factory=factory([1.0, 2.0, 3.0]),
        )

        # True (noiseless) value is zero. Averaging should ==> closer to zero.
        assert abs(zne_value_averaging) <= abs(zne_value_no_averaging)


def test_execute_with_zne_bad_arguments():
    """Tests errors are raised when execute_with_zne is called with bad
    arguments.
    """
    with pytest.raises(TypeError, match="Argument `factory` must be of type"):
        execute_with_zne(circ, executor, factory=RichardsonFactory)

    with pytest.raises(TypeError, match="Argument `scale_noise` must be"):
        execute_with_zne(circ, executor, scale_noise=None)


def test_error_zne_decorator():
    """Tests that the proper error is raised if the decorator is
    used without parenthesis.
    """
    with pytest.raises(TypeError, match="Decorator must be used with paren"):

        @zne_decorator
        def test_executor(circuit):
            return 0


def test_doc_is_preserved():
    """Tests that the doc of the original executor is preserved."""

    def first_executor(circuit):
        """Doc of the original executor."""
        return 0

    mit_executor = mitigate_executor(first_executor)
    assert mit_executor.__doc__ == first_executor.__doc__

    @zne_decorator()
    def second_executor(circuit):
        """Doc of the original executor."""
        return 0

    assert second_executor.__doc__ == first_executor.__doc__


def qiskit_measure(circuit, qid) -> qiskit.QuantumCircuit:
    """Helper function to measure one qubit."""
    # Ensure that we have a classical register of enough size available
    if len(circuit.clbits) == 0:
        reg = qiskit.ClassicalRegister(qid + 1, "cbits")
        circuit.add_register(reg)
    circuit.measure(0, qid)
    return circuit


def qiskit_executor(qp: QPROGRAM, shots: int = 10000) -> float:
    # initialize a qiskit noise model
    expectation = execute_with_shots_and_noise(
        qp,
        shots=shots,
        obs=ONE_QUBIT_GS_PROJECTOR,
        noise_model=initialized_depolarizing_noise(BASE_NOISE),
        seed=1,
    )
    return expectation


def get_counts(circuit: qiskit.QuantumCircuit):
    return AerSimulator().run(circuit, shots=100).result().get_counts()


def test_qiskit_execute_with_zne():
    true_zne_value = 1.0

    circuit = qiskit_measure(
        *generate_rb_circuits(
            n_qubits=1,
            num_cliffords=TEST_DEPTH,
            trials=1,
            return_type="qiskit",
        ),
        0,
    )
    base = qiskit_executor(circuit)
    zne_value = execute_with_zne(circuit, qiskit_executor)
    assert abs(true_zne_value - zne_value) < abs(true_zne_value - base)


@zne_decorator()
def qiskit_decorated_executor(qp: QPROGRAM) -> float:
    return qiskit_executor(qp)


def batched_qiskit_executor(circuits) -> List[float]:
    return [qiskit_executor(circuit) for circuit in circuits]


def test_qiskit_mitigate_executor():
    true_zne_value = 1.0

    circuit = qiskit_measure(
        *generate_rb_circuits(
            n_qubits=1,
            num_cliffords=TEST_DEPTH,
            trials=1,
            return_type="qiskit",
        ),
        0,
    )
    base = qiskit_executor(circuit)

    mitigated_executor = mitigate_executor(qiskit_executor)
    zne_value = mitigated_executor(circuit)
    assert abs(true_zne_value - zne_value) < abs(true_zne_value - base)
    batched_mitigated_executor = mitigate_executor(batched_qiskit_executor)
    batched_zne_values = batched_mitigated_executor([circuit] * 3)
    assert [
        abs(true_zne_value - batched_zne_value) < abs(true_zne_value - base)
        for batched_zne_value in batched_zne_values
    ]


def test_qiskit_zne_decorator():
    true_zne_value = 1.0

    circuit = qiskit_measure(
        *generate_rb_circuits(
            n_qubits=1,
            num_cliffords=TEST_DEPTH,
            trials=1,
            return_type="qiskit",
        ),
        0,
    )
    base = qiskit_executor(circuit)

    zne_value = qiskit_decorated_executor(circuit)
    assert abs(true_zne_value - zne_value) < abs(true_zne_value - base)


def test_qiskit_run_factory_with_number_of_shots():
    true_zne_value = 1.0

    scale_factors = [1.0, 3.0]
    shot_list = [10_000, 30_000]

    fac = inference.ExpFactory(
        scale_factors=scale_factors,
        shot_list=shot_list,
        asymptote=0.5,
    )

    circuit = qiskit_measure(
        *generate_rb_circuits(
            n_qubits=1,
            num_cliffords=TEST_DEPTH,
            trials=1,
            return_type="qiskit",
        ),
        0,
    )
    base = qiskit_executor(circuit)
    zne_value = fac.run(
        circuit,
        qiskit_executor,
        scale_noise=scaling.fold_gates_at_random,
    ).reduce()

    assert abs(true_zne_value - zne_value) < abs(true_zne_value - base)

    for i in range(len(fac._instack)):
        assert fac._instack[i] == {
            "scale_factor": scale_factors[i],
            "shots": shot_list[i],
        }


def test_qiskit_mitigate_executor_with_shot_list():
    true_zne_value = 1.0

    scale_factors = [1.0, 3.0]
    shot_list = [10_000, 30_000]

    fac = inference.ExpFactory(
        scale_factors=scale_factors,
        shot_list=shot_list,
        asymptote=0.5,
    )
    mitigated_executor = mitigate_executor(qiskit_executor, factory=fac)

    circuit = qiskit_measure(
        *generate_rb_circuits(
            n_qubits=1,
            num_cliffords=TEST_DEPTH,
            trials=1,
            return_type="qiskit",
        ),
        0,
    )
    base = qiskit_executor(circuit)
    zne_value = mitigated_executor(circuit)

    assert abs(true_zne_value - zne_value) < abs(true_zne_value - base)

    for i in range(len(fac._instack)):
        assert fac._instack[i] == {
            "scale_factor": scale_factors[i],
            "shots": shot_list[i],
        }


@pytest.mark.parametrize("order", [(0, 1), (1, 0), (0, 1, 2), (1, 2, 0)])
def test_qiskit_measurement_order_is_preserved_single_register(order):
    """Tests measurement order is preserved when folding, i.e., the dictionary
    of counts is the same as the original circuit on a noiseless simulator.
    """
    qreg, creg = (
        qiskit.QuantumRegister(len(order)),
        qiskit.ClassicalRegister(len(order)),
    )
    circuit = qiskit.QuantumCircuit(qreg, creg)

    circuit.x(qreg[0])
    for i in order:
        circuit.measure(qreg[i], creg[i])

    folded = scaling.fold_gates_at_random(circuit, scale_factor=1.0)

    assert get_counts(folded) == get_counts(circuit)


def test_qiskit_measurement_order_is_preserved_two_registers():
    """Tests measurement order is preserved when folding, i.e., the dictionary
    of counts is the same as the original circuit on a noiseless simulator.
    """
    n = 4
    qreg = qiskit.QuantumRegister(n)
    creg1, creg2 = (
        qiskit.ClassicalRegister(n // 2),
        qiskit.ClassicalRegister(n // 2),
    )
    circuit = qiskit.QuantumCircuit(qreg, creg1, creg2)

    circuit.x(qreg[0])
    circuit.x(qreg[2])

    # Some order of measurements.
    circuit.measure(qreg[0], creg2[1])
    circuit.measure(qreg[1], creg1[0])
    circuit.measure(qreg[2], creg1[1])
    circuit.measure(qreg[3], creg2[1])

    folded = scaling.fold_gates_at_random(circuit, scale_factor=1.0)

    assert get_counts(folded) == get_counts(circuit)


@pytest.mark.parametrize("circuit_type", SUPPORTED_PROGRAM_TYPES.keys())
def test_execute_with_zne_with_supported_circuits(circuit_type):
    # Define a circuit equivalent to the identity
    qreg = cirq.LineQubit.range(2)
    cirq_circuit = cirq.Circuit(
        cirq.H.on_each(qreg),
        cirq.CNOT(*qreg),
        cirq.CNOT(*qreg),
        cirq.H.on_each(qreg),
    )
    # Convert to one of the supported program types
    circuit = convert_from_mitiq(cirq_circuit, circuit_type)
    expected = generic_executor(circuit, noise_level=0.0)
    unmitigated = generic_executor(circuit)
    # Use odd scale factors for deterministic results
    fac = RichardsonFactory([1.0, 3.0, 5.0])
    zne_value = execute_with_zne(circuit, generic_executor, factory=fac)
    # Test zero noise limit is better than unmitigated expectation value
    assert abs(unmitigated - expected) > abs(zne_value - expected)


def test_layerwise_folding_with_zne():
    # Define a circuit equivalent to the identity
    qreg = cirq.LineQubit.range(2)
    circuit = cirq.Circuit(
        cirq.H.on_each(qreg),
        cirq.CNOT(*qreg),
        cirq.CNOT(*qreg),
        cirq.H.on_each(qreg),
    )
    circuit_depth = len(circuit)
    mock_executor = Mock(side_effect=lambda _: random.random())
    layer_to_fold = 0
    fold_layer_func = get_layer_folding(layer_to_fold)
    scale_factors = [1, 3, 5]
    factory = RichardsonFactory(scale_factors)

    execute_with_zne(
        circuit, mock_executor, factory=factory, scale_noise=fold_layer_func
    )
    assert mock_executor.call_count == len(scale_factors)
    circuit_depths = [
        len(args[0]) for args, kwargs in mock_executor.call_args_list
    ]
    assert circuit_depths == [
        circuit_depth,
        circuit_depth + 2,
        circuit_depth + 4,
    ]


def test_execute_with_zne_transpiled_qiskit_circuit():
    """Tests ZNE when transpiling to a Qiskit device. Note transpiling can
    introduce idle (unused) qubits to the circuit.
    """
    from qiskit_ibm_runtime.fake_provider import FakeSantiagoV2

    santiago = FakeSantiagoV2()
    backend = AerSimulator.from_backend(santiago)

    def execute(circuit: qiskit.QuantumCircuit, shots: int = 8192) -> float:
        job = backend.run(circuit, shots=shots)
        return job.result().get_counts().get("00", 0.0) / shots

    qreg = qiskit.QuantumRegister(2)
    creg = qiskit.ClassicalRegister(2)
    circuit = qiskit.QuantumCircuit(qreg, creg)
    for _ in range(10):
        circuit.x(qreg)

    circuit.measure(qreg, creg)
    circuit = qiskit.transpile(circuit, backend, optimization_level=0)

    true_value = 1.0
    zne_value = execute_with_zne(circuit, execute)

    # Note: Unmitigated value is also (usually) within 10% of the true value.
    # This is more to test usage than effectiveness.
    assert abs(zne_value - true_value) < 0.1


def test_execute_zne_on_qiskit_circuit_with_QFT():
    """Tests ZNE of a Qiskit device with a QFT gate."""

    def qs_noisy_simulation(
        circuit: qiskit.QuantumCircuit, shots: int = 1
    ) -> float:
        noise_model = initialized_depolarizing_noise(noise_level=0.02)
        backend = AerSimulator(noise_model=noise_model)
        job = backend.run(circuit.decompose(), shots=shots)
        return job.result().get_counts().get("0", 0.0) / shots

    circuit = qiskit.QuantumCircuit(1)
    circuit &= qiskit.circuit.library.QFT(1)
    circuit.measure_all()

    mitigated = execute_with_zne(circuit, qs_noisy_simulation)
    assert abs(mitigated) < 1000


@pytest.mark.parametrize(
    "noise_scaling_method",
    [fold_gates_at_random, insert_id_layers, fold_global, fold_all],
)
@pytest.mark.parametrize(
    "extrapolation_factory", [RichardsonFactory, LinearFactory]
)
@pytest.mark.parametrize(
    "conversion_func",
    [None, to_qiskit, to_braket, to_pennylane, to_pyquil, to_qibo],
)
def test_two_stage_zne(
    noise_scaling_method, extrapolation_factory, conversion_func
):
    qreg = cirq.LineQubit.range(2)
    cirq_circuit = cirq.Circuit(
        cirq.H.on_each(qreg),
        cirq.CNOT(*qreg),
        cirq.CNOT(*qreg),
        cirq.H.on_each(qreg),
    )
    if conversion_func is not None:
        frontend_circuit = conversion_func(cirq_circuit)
    else:
        frontend_circuit = cirq_circuit

    scale_factors = [1, 3, 5]
    circs = construct_circuits(
        frontend_circuit, scale_factors, noise_scaling_method
    )

    assert len(circs) == len(scale_factors)

    np.random.seed(42)

    def executor(circuit):
        return np.random.random()

    results = [executor(cirq_circuit) for _ in range(3)]
    extrapolation_method = extrapolation_factory.extrapolate
    two_stage_zne_res = combine_results(
        scale_factors, results, extrapolation_method
    )

    assert isinstance(two_stage_zne_res, float)

    np.random.seed(42)
    zne_res = execute_with_zne(
        cirq_circuit,
        executor,
        factory=extrapolation_factory(scale_factors),
        scale_noise=noise_scaling_method,
    )
    assert np.isclose(zne_res, two_stage_zne_res)


def test_default_scaling_option_two_stage_zne():
    qreg = cirq.LineQubit.range(2)
    cirq_circuit = cirq.Circuit(
        cirq.H.on_each(qreg),
        cirq.CNOT(*qreg),
        cirq.CNOT(*qreg),
        cirq.H.on_each(qreg),
    )

    scale_factors = [3, 5, 6, 8]

    circs_default_scaling_method = construct_circuits(
        cirq_circuit, scale_factors
    )

    for i in range(len(scale_factors)):
        assert len(circs_default_scaling_method[i]) > len(cirq_circuit)



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Unit tests for scaling noise by unitary folding."""

import numpy as np
import pytest
from cirq import (
    Circuit,
    GridQubit,
    InsertStrategy,
    LineQubit,
    equal_up_to_global_phase,
    inverse,
    ops,
    ry,
    rz,
    testing,
)
from pyquil import Program, gates
from pyquil.quilbase import Pragma
from qiskit import ClassicalRegister, QuantumCircuit, QuantumRegister
from sympy import Symbol

from mitiq.interface import (
    CircuitConversionError,
    convert_from_mitiq,
    convert_to_mitiq,
)
from mitiq.utils import _equal
from mitiq.zne.scaling.folding import (
    UnfoldableCircuitError,
    _apply_fold_mask,
    _create_fold_mask,
    _create_weight_mask,
    _default_weight,
    _fold_all,
    _squash_moments,
    fold_all,
    fold_gates_at_random,
    fold_global,
)


def test_squash_moments_two_qubits():
    """Tests squashing moments in a two-qubit circuit with 'staggered' single
    qubit gates.
    """
    # Test circuit:
    # 0: ───────H───────H───────H───────H───────H───
    #
    # 1: ───H───────H───────H───────H───────H───────

    # Get the test circuit
    d = 10
    qreg = LineQubit.range(2)
    circuit = Circuit()
    for i in range(d):
        circuit.insert(0, ops.H.on(qreg[i % 2]), strategy=InsertStrategy.NEW)
    assert len(circuit) == d

    # Squash the moments
    squashed = _squash_moments(circuit)
    assert len(squashed) == d // 2


def test_squash_moments_returns_new_circuit_and_doesnt_modify_input_circuit():
    """Tests that squash moments returns a new circuit and doesn't modify the
    input circuit.
    """
    qbit = GridQubit(0, 0)
    circ = Circuit(ops.H.on(qbit))
    squashed = _squash_moments(circ)
    assert len(squashed) == 1
    assert circ is not squashed
    assert _equal(circ, Circuit(ops.H.on(qbit)))


def test_squash_moments_never_increases_moments():
    """Squashes moments for several random circuits and ensures the squashed
    circuit always <= # moments as the input circuit.
    """
    for _ in range(50):
        circuit = testing.random_circuit(
            qubits=5, n_moments=8, op_density=0.75
        )
        squashed = _squash_moments(circuit)
        assert len(squashed) <= len(circuit)


@pytest.mark.parametrize("scale_factor", (1, 3, 5))
@pytest.mark.parametrize("with_measurements", (True, False))
def test_fold_all(scale_factor, with_measurements):
    # Test circuit
    # 0: ───H───@───@───
    #           │   │
    # 1: ───H───X───@───
    #               │
    # 2: ───H───X───X───
    qreg = LineQubit.range(3)
    circ = Circuit(
        [ops.H.on_each(*qreg)],
        [ops.CNOT.on(qreg[0], qreg[1])],
        [ops.X.on(qreg[2])],
        [ops.TOFFOLI.on(*qreg)],
    )
    if with_measurements:
        circ.append(ops.measure_each(*qreg))

    folded = fold_all(circ, scale_factor=scale_factor)

    correct = Circuit(
        [ops.H.on_each(*qreg)] * scale_factor,
        [ops.CNOT.on(qreg[0], qreg[1])] * scale_factor,
        [ops.X.on(qreg[2])] * scale_factor,
        [ops.TOFFOLI.on(*qreg)] * scale_factor,
    )
    if with_measurements:
        correct.append(ops.measure_each(*qreg))
    assert _equal(folded, correct, require_qubit_equality=True)


def test_fold_all_exclude_with_gates():
    circuit = Circuit(
        [ops.H(LineQubit(0))],
        [ops.CNOT(*LineQubit.range(2))],
        [ops.TOFFOLI(*LineQubit.range(3))],
    )

    folded = fold_all(circuit, scale_factor=3.0, exclude={ops.H})
    correct = Circuit(
        [ops.H(LineQubit(0))],
        [ops.CNOT(*LineQubit.range(2))] * 3,
        [ops.TOFFOLI(*LineQubit.range(3))] * 3,
    )
    assert _equal(folded, correct, require_qubit_equality=True)

    folded = fold_all(circuit, scale_factor=3.0, exclude={ops.CNOT})
    correct = Circuit(
        [ops.H(LineQubit(0))] * 3,
        [ops.CNOT(*LineQubit.range(2))],
        [ops.TOFFOLI(*LineQubit.range(3))] * 3,
    )
    assert _equal(folded, correct, require_qubit_equality=True)

    folded = fold_all(circuit, scale_factor=5.0, exclude={ops.H, ops.TOFFOLI})
    correct = Circuit(
        [ops.H(LineQubit(0))],
        [ops.CNOT(*LineQubit.range(2))] * 5,
        [ops.TOFFOLI(*LineQubit.range(3))],
    )
    assert _equal(folded, correct, require_qubit_equality=True)


def test_fold_all_exclude_with_strings():
    circuit = Circuit(
        [ops.H(LineQubit(0))],
        [ops.CNOT(*LineQubit.range(2))],
        [ops.TOFFOLI(*LineQubit.range(3))],
    )

    folded = fold_all(circuit, scale_factor=3.0, exclude={"single"})
    correct = Circuit(
        [ops.H(LineQubit(0))],
        [ops.CNOT(*LineQubit.range(2))] * 3,
        [ops.TOFFOLI(*LineQubit.range(3))] * 3,
    )
    assert _equal(folded, correct, require_qubit_equality=True)

    folded = fold_all(circuit, scale_factor=3.0, exclude={"double"})
    correct = Circuit(
        [ops.H(LineQubit(0))] * 3,
        [ops.CNOT(*LineQubit.range(2))],
        [ops.TOFFOLI(*LineQubit.range(3))] * 3,
    )
    assert _equal(folded, correct, require_qubit_equality=True)

    folded = fold_all(circuit, scale_factor=5.0, exclude={"single", "triple"})
    correct = Circuit(
        [ops.H(LineQubit(0))],
        [ops.CNOT(*LineQubit.range(2))] * 5,
        [ops.TOFFOLI(*LineQubit.range(3))],
    )
    assert _equal(folded, correct, require_qubit_equality=True)


@pytest.mark.parametrize("skip", (frozenset((0, 1)), frozenset((0, 3, 7))))
def test_fold_all_skip_moments(skip):
    circuit = testing.random_circuit(
        qubits=3,
        n_moments=7,
        op_density=1,
        random_state=1,
        gate_domain={ops.H: 1, ops.X: 1, ops.CNOT: 2},
    )
    folded = _fold_all(circuit, skip_moments=skip)

    correct = Circuit()
    for i, moment in enumerate(circuit):
        times_to_add = 3 * (i not in skip) + (i in skip)
        for _ in range(times_to_add):
            correct += moment
    assert _equal(
        _squash_moments(folded),
        _squash_moments(correct),
        require_qubit_equality=True,
    )


def test_folding_with_bad_scale_factor():
    for fold_function in (
        fold_all,
        fold_gates_at_random,
    ):
        with pytest.raises(ValueError, match="Requires scale_factor >= 1"):
            fold_function(Circuit(), scale_factor=0.0)


def test_create_mask_with_bad_scale_factor():
    with pytest.raises(ValueError, match="Requires scale_factor >= 1"):
        _create_fold_mask([1], scale_factor=0.999)


def test_fold_all_bad_exclude():
    with pytest.raises(ValueError, match="Do not know how to parse"):
        fold_all(
            Circuit(),
            scale_factor=1.0,
            exclude=frozenset(("not a gate name",)),
        )

    with pytest.raises(ValueError, match="Do not know how to exclude"):
        fold_all(Circuit(), scale_factor=1.0, exclude=frozenset((7,)))


@pytest.mark.parametrize(
    "fold_method",
    [
        fold_gates_at_random,
        fold_global,
    ],
)
def test_fold_with_intermediate_measurements_raises_error(fold_method):
    """Tests local folding functions raise an error on circuits with
    intermediate measurements.
    """
    qbit = LineQubit(0)
    circ = Circuit([ops.H.on(qbit)], [ops.measure(qbit)], [ops.T.on(qbit)])
    with pytest.raises(
        UnfoldableCircuitError,
        match="Circuit contains intermediate measurements",
    ):
        fold_method(circ, scale_factor=3.0)


@pytest.mark.parametrize(
    "fold_method",
    [
        fold_gates_at_random,
        fold_global,
    ],
)
def test_fold_with_channels_raises_error(fold_method):
    """Tests local folding functions raise an error on circuits with
    non-invertible channels (which are not measurements).
    """
    qbit = LineQubit(0)
    circ = Circuit(
        ops.H.on(qbit), ops.depolarize(p=0.1).on(qbit), ops.measure(qbit)
    )
    with pytest.raises(
        UnfoldableCircuitError, match="Circuit contains non-invertible"
    ):
        fold_method(circ, scale_factor=3.0)


@pytest.mark.parametrize(
    "fold_method",
    [
        fold_gates_at_random,
        fold_global,
    ],
)
def test_parametrized_circuit_folding(fold_method):
    """Checks if the circuit is folded as expected when the circuit operations
    have a valid inverse.
    """
    theta = Symbol("theta")
    q = LineQubit(0)
    ansatz_circ = Circuit(ry(theta).on(q))
    folded_circ = fold_method(ansatz_circ, scale_factor=3.0)
    expected_circ = Circuit(ry(theta).on(q), ry(-theta).on(q), ry(theta).on(q))
    assert _equal(folded_circ, expected_circ)


@pytest.mark.parametrize(
    "fold_method",
    [
        fold_gates_at_random,
        fold_global,
    ],
)
def test_parametrized_circuit_folding_terminal_measurement(fold_method):
    """Checks if the circuit with a terminal measurement is folded as expected
    when the circuit operations have a valid inverse.
    """
    theta = Symbol("theta")
    q = LineQubit(0)
    ansatz_circ = Circuit(ry(theta).on(q), ops.measure(q))
    folded_circ = fold_method(ansatz_circ, scale_factor=3.0)
    expected_circ = Circuit(
        ry(theta).on(q), ry(-theta).on(q), ry(theta).on(q), ops.measure(q)
    )
    assert _equal(folded_circ, expected_circ)


@pytest.mark.parametrize(
    "fold_method",
    [
        fold_gates_at_random,
        fold_global,
    ],
)
def test_errors_raised_parametrized_circuits(fold_method):
    """Checks if proper error is raised in a symbolic circuit when it cannot
    be folded.
    """
    theta = Symbol("theta")
    q = LineQubit(0)
    ansatz_circ = Circuit(ry(theta).on(q), ops.measure(q), rz(theta).on(q))
    with pytest.raises(
        UnfoldableCircuitError,
        match="Circuit contains intermediate measurements",
    ):
        fold_method(ansatz_circ, scale_factor=3.0)

    qbit = LineQubit(0)
    circ = Circuit(
        ry(theta).on(q), ops.depolarize(p=0.1).on(qbit), ops.measure(qbit)
    )
    with pytest.raises(
        UnfoldableCircuitError, match="Circuit contains non-invertible"
    ):
        fold_method(circ, scale_factor=3.0)


def test_fold_gates_at_random_no_stretch():
    """Tests folded circuit is identical for a scale factor of one."""
    circuit = testing.random_circuit(qubits=3, n_moments=10, op_density=0.99)
    folded = fold_gates_at_random(circuit, scale_factor=1, seed=None)
    assert _equal(folded, _squash_moments(circuit))


def test_fold_gates_at_random_seed_one_qubit():
    """Test for folding gates at random on a one qubit circuit with a seed for
    repeated behavior.
    """
    qubit = LineQubit(0)
    circuit = Circuit([ops.X.on(qubit), ops.Y.on(qubit), ops.Z.on(qubit)])
    # Small scale
    folded = fold_gates_at_random(circuit, scale_factor=1.4, seed=3)
    correct = Circuit(
        [ops.X.on(qubit)], [ops.Y.on(qubit)] * 3, [ops.Z.on(qubit)]
    )
    assert _equal(folded, correct)

    # Medium scale, fold two gates
    folded = fold_gates_at_random(circuit, scale_factor=2.5, seed=2)
    correct = Circuit(
        [ops.X.on(qubit)],
        [ops.Y.on(qubit)] * 3,
        [ops.Z.on(qubit)] * 3,
    )
    assert _equal(folded, correct)

    # Max scale, fold three gates
    folded = fold_gates_at_random(circuit, scale_factor=3, seed=3)
    correct = Circuit(
        [ops.X.on(qubit)] * 3,
        [ops.Y.on(qubit)] * 3,
        [ops.Z.on(qubit)] * 3,
    )
    assert _equal(folded, correct)


def test_fold_random_min_stretch():
    """Tests that folding at random with min scale returns a copy of the
    input circuit.
    """
    # Test circuit
    # 0: ───H───@───@───
    #           │   │
    # 1: ───H───X───@───
    #               │
    # 2: ───H───T───X───
    qreg = LineQubit.range(3)
    circ = Circuit(
        [ops.H.on_each(*qreg)],
        [ops.CNOT.on(qreg[0], qreg[1])],
        [ops.T.on(qreg[2])],
        [ops.TOFFOLI.on(*qreg)],
    )

    folded = fold_gates_at_random(circ, scale_factor=1, seed=1)
    assert _equal(folded, circ)
    assert folded is not circ


def test_fold_random_max_stretch():
    """Tests that folding at random with max scale folds all gates on a
    multi-qubit circuit.
    """
    # Test circuit
    # 0: ───H───@───@───
    #           │   │
    # 1: ───H───X───@───
    #               │
    # 2: ───H───T───X───
    qreg = LineQubit.range(3)
    circ = Circuit(
        [ops.H.on_each(*qreg)],
        [ops.CNOT.on(qreg[0], qreg[1])],
        [ops.T.on(qreg[2])],
        [ops.TOFFOLI.on(*qreg)],
    )

    folded = fold_gates_at_random(circ, scale_factor=3, seed=1)
    correct = Circuit(
        [ops.H.on_each(*qreg)] * 3,
        [ops.CNOT.on(qreg[0], qreg[1])] * 3,
        [ops.T.on(qreg[2]), ops.T.on(qreg[2]) ** -1, ops.T.on(qreg[2])],
        [ops.TOFFOLI.on(*qreg)] * 3,
    )
    assert _equal(folded, correct)


def test_fold_random_scale_factor_larger_than_three():
    """Folds at random with a scale_factor larger than three."""
    qreg = LineQubit.range(2)
    circuit = Circuit([ops.SWAP.on(*qreg)], [ops.CNOT.on(*qreg)])
    folded = fold_gates_at_random(circuit, scale_factor=6.0, seed=0)
    correct = Circuit([ops.SWAP.on(*qreg)] * 5, [ops.CNOT.on(*qreg)] * 7)
    assert len(folded) == 12
    assert _equal(folded, correct)


def test_fold_random_no_repeats():
    """Tests folding at random to ensure that no gates are folded twice and
    folded gates are not folded again.
    """
    # Test circuit:
    # 0: ───H───@───Y───@───
    #           │       │
    # 1: ───────X───X───@───
    # Note that each gate only occurs once and is self-inverse.
    # This allows us to check that no gates are folded more than once
    qreg = LineQubit.range(2)
    circ = Circuit(
        [ops.H.on_each(qreg[0])],
        [ops.CNOT.on(*qreg)],
        [ops.X.on(qreg[1])],
        [ops.Y.on(qreg[0])],
        [ops.CZ.on(*qreg)],
    )
    circuit_ops = set(circ.all_operations())

    for scale in np.linspace(1.0, 3.0, 5):
        folded = fold_gates_at_random(circ, scale_factor=scale, seed=1)
        gates = list(folded.all_operations())
        counts = {gate: gates.count(gate) for gate in circuit_ops}
        assert all(count <= 3 for count in counts.values())


def test_fold_random_with_terminal_measurements_min_stretch():
    """Tests folding from left with terminal measurements."""
    # Test circuit
    # 0: ───H───@───@───M───
    #           │   │
    # 1: ───H───X───@───M───
    #               │
    # 2: ───H───T───X───M───
    qreg = LineQubit.range(3)
    circ = Circuit(
        [ops.H.on_each(*qreg)],
        [ops.CNOT.on(qreg[0], qreg[1])],
        [ops.T.on(qreg[2])],
        [ops.TOFFOLI.on(*qreg)],
        [ops.measure_each(*qreg)],
    )
    folded = fold_gates_at_random(circ, scale_factor=1.0)
    correct = Circuit(
        [ops.H.on_each(*qreg)],
        [ops.CNOT.on(qreg[0], qreg[1])],
        [ops.T.on(qreg[2])],
        [ops.TOFFOLI.on(*qreg)],
        [ops.measure_each(*qreg)],
    )
    assert _equal(folded, correct)


def test_fold_random_with_terminal_measurements_max_stretch():
    """Tests folding from left with terminal measurements."""
    # Test circuit
    # 0: ───H───@───@───M───
    #           │   │
    # 1: ───H───X───@───M───
    #               │
    # 2: ───H───T───X───M───
    qreg = LineQubit.range(3)
    circ = Circuit(
        [ops.H.on_each(*qreg)],
        [ops.CNOT.on(qreg[0], qreg[1])],
        [ops.T.on(qreg[2])],
        [ops.TOFFOLI.on(*qreg)],
        [ops.measure_each(*qreg)],
    )
    folded = fold_gates_at_random(circ, scale_factor=3.0)
    correct = Circuit(
        [ops.H.on_each(*qreg)] * 3,
        [ops.CNOT.on(qreg[0], qreg[1])] * 3,
        [ops.T.on(qreg[2]), ops.T.on(qreg[2]) ** -1, ops.T.on(qreg[2])],
        [ops.TOFFOLI.on(*qreg)] * 3,
        [ops.measure_each(*qreg)],
    )
    assert _equal(folded, correct)


def test_global_fold_min_stretch():
    """Tests that global fold with scale = 1 is the same circuit."""
    # Test circuit
    # 0: ───H───@───@───
    #           │   │
    # 1: ───H───X───@───
    #               │
    # 2: ───H───T───X───
    qreg = LineQubit.range(3)
    circ = Circuit(
        [ops.H.on_each(*qreg)],
        [ops.CNOT.on(qreg[0], qreg[1])],
        [ops.T.on(qreg[2])],
        [ops.TOFFOLI.on(*qreg)],
    )

    folded = fold_global(circ, 1.0)
    assert _equal(folded, circ)
    assert folded is not circ


def test_global_fold_min_stretch_with_terminal_measurements():
    """Tests that global fold with scale = 1 is the same circuit."""
    # Test circuit
    # 0: ───H───@───@───M───
    #           │   │
    # 1: ───H───X───@───M───
    #               │
    # 2: ───H───T───X───M───
    qreg = LineQubit.range(3)
    circ = Circuit(
        [ops.H.on_each(*qreg)],
        [ops.CNOT.on(qreg[0], qreg[1])],
        [ops.T.on(qreg[2])],
        [ops.TOFFOLI.on(*qreg)],
        [ops.measure_each(*qreg)],
    )
    folded = fold_global(circ, scale_factor=1.0)
    assert _equal(folded, circ)
    assert folded is not circ


def test_global_fold_stretch_factor_of_three():
    """Tests global folding with the scale as a factor of 3."""
    # Test circuit
    # 0: ───H───@───@───
    #           │   │
    # 1: ───H───X───@───
    #               │
    # 2: ───H───T───X───
    qreg = LineQubit.range(3)
    circ = Circuit(
        [ops.H.on_each(*qreg)],
        [ops.CNOT.on(qreg[0], qreg[1])],
        [ops.T.on(qreg[2])],
        [ops.TOFFOLI.on(*qreg)],
    )
    folded = fold_global(circ, scale_factor=3.0)
    correct = Circuit(circ, inverse(circ), circ)
    assert _equal(folded, correct)


def test_global_fold_stretch_factor_of_three_with_terminal_measurements():
    """Tests global folding with the scale as a factor of 3 for a circuit
    with terminal measurements.
    """
    # Test circuit
    # 0: ───H───@───@───M───
    #           │   │
    # 1: ───H───X───@───M───
    #               │
    # 2: ───H───T───X───M───
    qreg = LineQubit.range(3)
    circ = Circuit(
        [ops.H.on_each(*qreg)],
        [ops.CNOT.on(qreg[0], qreg[1])],
        [ops.T.on(qreg[2])],
        [ops.TOFFOLI.on(*qreg)],
    )
    meas = Circuit([ops.measure_each(*qreg)])
    folded = fold_global(circ + meas, scale_factor=3.0)
    correct = Circuit(circ, inverse(circ), circ, meas)
    assert _equal(folded, correct)
    # Test the number of moments too
    assert len(folded) == len(correct)


def test_global_fold_stretch_factor_nine_with_terminal_measurements():
    """Tests global folding with the scale as a factor of 9 for a circuit
    with terminal measurements.
    """
    # Test circuit
    # 0: ───H───@───@───M───
    #           │   │
    # 1: ───H───X───@───M───
    #               │
    # 2: ───H───T───X───M───
    qreg = LineQubit.range(3)
    circ = Circuit(
        [ops.H.on_each(*qreg)],
        [ops.CNOT.on(qreg[0], qreg[1])],
        [ops.T.on(qreg[2])],
        [ops.TOFFOLI.on(*qreg)],
    )
    meas = Circuit([ops.measure_each(*qreg)])
    folded = fold_global(circ + meas, scale_factor=9.0)
    correct = Circuit([circ, inverse(circ)] * 4, [circ], [meas])
    assert _equal(folded, correct)


def test_global_fold_stretch_factor_eight_terminal_measurements():
    """Tests global folding with a scale factor not a multiple of three so
    that local folding is also called.
    """
    # Test circuit
    # 0: ───H───@───@───M───
    #           │   │
    # 1: ───H───X───@───M───
    #               │
    # 2: ───H───T───X───M───
    qreg = LineQubit.range(3)
    circ = Circuit(
        [ops.H.on_each(*qreg)],
        [ops.CNOT.on(qreg[0], qreg[1])],
        [ops.T.on(qreg[2])],
        [ops.TOFFOLI.on(*qreg)],
    )
    meas = Circuit(ops.measure_each(*qreg))
    folded = fold_global(circ + meas, scale_factor=3.5)
    correct = Circuit(
        circ,
        inverse(circ),
        circ,
        inverse(
            Circuit(
                [ops.T.on(qreg[2])],
                [ops.TOFFOLI.on(*qreg)],
            )
        ),
        [ops.T.on(qreg[2])],
        [ops.TOFFOLI.on(*qreg)],
        meas,
    )
    assert _equal(folded, correct)


def test_global_fold_moment_structure_maintained_full_scale_factors():
    """Tests global folding maintains the input circuit's moment structure."""
    # Test circuit 1
    # 0: ───H───────────────

    # 1: ───────Z───────────

    # 2: ───────────S───────

    # 3: ───────────────T───

    qreg = LineQubit.range(4)

    gate_list1 = [ops.H, ops.Z, ops.S, ops.T]
    circuit1 = Circuit(gate_list1[0](qreg[0]))

    for i in range(1, 4):
        circuit1 += Circuit(gate_list1[i](qreg[i]))
    folded = fold_global(circuit1, scale_factor=3)
    correct = Circuit(
        circuit1,
        inverse(circuit1),
        circuit1,
    )
    assert _equal(folded, correct)

    # Test Circuit 2
    # 0: ───H───@───────@───
    #           │       |
    # 1: ───H───X───────@───
    #                   │
    # 2: ───H───────T───X───
    qreg = LineQubit.range(3)
    gate_list = [
        ops.CNOT.on(qreg[0], qreg[1]),
        [ops.T.on(qreg[2])],
        [ops.TOFFOLI.on(*qreg)],
    ]
    circ = Circuit([ops.H.on_each(*qreg)])
    for i in range(len(gate_list)):
        circ += Circuit(gate_list[i])
    folded = fold_global(circ, scale_factor=3)
    correct = Circuit(
        circ,
        inverse(circ),
        circ,
    )
    assert _equal(folded, correct)


def test_global_fold_moment_structure_maintained_partial_scale_factors():
    """Tests global folding maintains the input circuit's moment structure."""
    # Test circuit 1
    # 0: ───H───────────────

    # 1: ───────Z───────────

    # 2: ───────────S───────

    # 3: ───────────────T───

    qreg = LineQubit.range(4)

    gate_list1 = [ops.H, ops.Z, ops.S, ops.T]
    circuit1 = Circuit(gate_list1[0](qreg[0]))

    for i in range(1, 4):
        circuit1 += Circuit(gate_list1[i](qreg[i]))
    folded1 = fold_global(circuit1, scale_factor=1.5)
    correct1 = Circuit(circuit1, inverse(circuit1)[0], circuit1[-1])
    assert _equal(folded1, correct1)

    folded2 = fold_global(circuit1, scale_factor=2.5)
    correct2 = Circuit(circuit1, inverse(circuit1)[0:3], circuit1[1:])
    assert _equal(folded2, correct2)

    folded3 = fold_global(circuit1, scale_factor=2.75)
    correct3 = Circuit(circuit1, inverse(circuit1), circuit1)
    assert _equal(folded3, correct3)


def test_convert_to_from_mitiq_qiskit():
    """Basic test for converting a Qiskit circuit to a Cirq circuit."""
    # Test Qiskit circuit:
    #          ┌───┐
    # q0_0: |0>┤ H ├──■──
    #          └───┘┌─┴─┐
    # q0_1: |0>─────┤ X ├
    #               └───┘
    qiskit_qreg = QuantumRegister(2)
    qiskit_circuit = QuantumCircuit(qiskit_qreg)
    qiskit_circuit.h(qiskit_qreg[0])
    qiskit_circuit.cx(*qiskit_qreg)

    # Convert to a mitiq circuit
    mitiq_circuit, input_circuit_type = convert_to_mitiq(qiskit_circuit)
    assert isinstance(mitiq_circuit, Circuit)

    # Check correctness
    mitiq_qreg = LineQubit.range(2)
    correct_mitiq_circuit = Circuit(
        ops.H.on(mitiq_qreg[0]), ops.CNOT.on(*mitiq_qreg)
    )
    assert _equal(
        mitiq_circuit, correct_mitiq_circuit, require_qubit_equality=False
    )

    # Convert back to original circuit type
    original_circuit = convert_from_mitiq(mitiq_circuit, input_circuit_type)
    assert isinstance(original_circuit, QuantumCircuit)


def test_fold_at_random_with_qiskit_circuits():
    """Tests folding at random with Qiskit circuits."""
    # Test Qiskit circuit:
    #          ┌───┐
    # q0_0: |0>┤ H ├──■────■──
    #          ├───┤┌─┴─┐  │
    # q0_1: |0>┤ H ├┤ X ├──■──
    #          ├───┤├───┤┌─┴─┐
    # q0_2: |0>┤ H ├┤ T ├┤ X ├
    #          └───┘└───┘└───┘
    qiskit_qreg = QuantumRegister(3)
    qiskit_creg = ClassicalRegister(3)
    qiskit_circuit = QuantumCircuit(qiskit_qreg, qiskit_creg)
    qiskit_circuit.h(qiskit_qreg)
    qiskit_circuit.cx(qiskit_qreg[0], qiskit_qreg[1])
    qiskit_circuit.t(qiskit_qreg[2])
    qiskit_circuit.ccx(*qiskit_qreg)
    qiskit_circuit.measure(qiskit_qreg, qiskit_creg)

    folded_circuit = fold_gates_at_random(
        qiskit_circuit, scale_factor=1.0, return_mitiq=True
    )

    qreg = LineQubit.range(3)
    correct_folded_circuit = Circuit(
        [ops.H.on_each(*qreg)],
        [ops.CNOT.on(qreg[0], qreg[1])],
        [ops.T.on(qreg[2])],
        [ops.TOFFOLI.on(*qreg)],
        [ops.measure_each(*qreg)],
    )

    assert isinstance(folded_circuit, Circuit)
    assert _equal(folded_circuit, correct_folded_circuit)

    # Keep the input type
    qiskit_folded_circuit = fold_gates_at_random(
        qiskit_circuit, scale_factor=1.0
    )
    assert isinstance(qiskit_folded_circuit, QuantumCircuit)
    assert qiskit_folded_circuit.qregs == qiskit_circuit.qregs
    assert qiskit_folded_circuit.cregs == qiskit_circuit.cregs


def test_fold_global_with_qiskit_circuits():
    """Tests _fold_local with input Qiskit circuits."""
    # Test Qiskit circuit:
    #          ┌───┐
    # q0_0: |0>┤ H ├──■────■──
    #          ├───┤┌─┴─┐  │
    # q0_1: |0>┤ H ├┤ X ├──■──
    #          ├───┤├───┤┌─┴─┐
    # q0_2: |0>┤ H ├┤ T ├┤ X ├
    #          └───┘└───┘└───┘
    qiskit_qreg = QuantumRegister(3)
    qiskit_creg = ClassicalRegister(3)
    qiskit_circuit = QuantumCircuit(qiskit_qreg, qiskit_creg)
    qiskit_circuit.h(qiskit_qreg)
    qiskit_circuit.cx(qiskit_qreg[0], qiskit_qreg[1])
    qiskit_circuit.t(qiskit_qreg[2])
    qiskit_circuit.ccx(*qiskit_qreg)
    qiskit_circuit.measure(qiskit_qreg, qiskit_creg)

    # Return mitiq circuit
    folded_circuit = fold_global(
        qiskit_circuit,
        scale_factor=2.71828,
        fold_method=fold_gates_at_random,
        return_mitiq=True,
    )
    assert isinstance(folded_circuit, Circuit)

    # Return input circuit type
    folded_qiskit_circuit = fold_global(
        qiskit_circuit, scale_factor=2.0, fold_method=fold_gates_at_random
    )
    assert isinstance(folded_qiskit_circuit, QuantumCircuit)
    assert folded_qiskit_circuit.qregs == qiskit_circuit.qregs
    assert folded_qiskit_circuit.cregs == qiskit_circuit.cregs


def test_fold_global_with_qiskit_circuits_and_idle_qubits():
    """Tests _fold_local with input Qiskit circuits where idle qubits are
    interspered.
    """
    # Test Qiskit circuit:
    #           ┌───┐          ┌─┐
    #  q4_0: |0>┤ H ├──■────■──┤M├──────
    #           └───┘  │    │  └╥┘
    #  q4_1: |0>───────┼────┼───╫───────
    #           ┌───┐┌─┴─┐  │   ║ ┌─┐
    #  q4_2: |0>┤ H ├┤ X ├──■───╫─┤M├───
    #           └───┘└───┘  │   ║ └╥┘
    #  q4_3: |0>────────────┼───╫──╫────
    #           ┌───┐┌───┐┌─┴─┐ ║  ║ ┌─┐
    #  q4_4: |0>┤ H ├┤ T ├┤ X ├─╫──╫─┤M├
    #           └───┘└───┘└───┘ ║  ║ └╥┘
    #  c4:    5/════════════════╩══╩══╩═
    #                           0  2  4
    qiskit_qreg = QuantumRegister(5)
    qiskit_creg = ClassicalRegister(5)
    qiskit_circuit = QuantumCircuit(qiskit_qreg, qiskit_creg)
    qiskit_circuit.h(qiskit_qreg[0])
    qiskit_circuit.h(qiskit_qreg[2])
    qiskit_circuit.h(qiskit_qreg[4])
    qiskit_circuit.cx(qiskit_qreg[0], qiskit_qreg[2])
    qiskit_circuit.t(qiskit_qreg[4])
    qiskit_circuit.ccx(qiskit_qreg[0], qiskit_qreg[2], qiskit_qreg[4])
    qiskit_circuit.measure(qiskit_qreg[0], qiskit_creg[0])
    qiskit_circuit.measure(qiskit_qreg[2], qiskit_creg[2])
    qiskit_circuit.measure(qiskit_qreg[4], qiskit_creg[4])

    # Return mitiq circuit
    folded_circuit = fold_global(
        qiskit_circuit,
        scale_factor=2.71828,
        fold_method=fold_gates_at_random,
        return_mitiq=True,
    )
    assert isinstance(folded_circuit, Circuit)

    # Return input circuit type
    folded_qiskit_circuit = fold_global(
        qiskit_circuit, scale_factor=2.0, fold_method=fold_gates_at_random
    )
    assert isinstance(folded_qiskit_circuit, QuantumCircuit)
    assert folded_qiskit_circuit.qregs == qiskit_circuit.qregs
    assert folded_qiskit_circuit.cregs == qiskit_circuit.cregs


def test_fold_left_squash_moments():
    """Tests folding from left with kwarg squash_moments."""
    # Test circuit
    # 0: ───H───@───@───M───
    #           │   │
    # 1: ───H───X───@───M───
    #               │
    # 2: ───H───T───X───M───
    qreg = LineQubit.range(3)
    circ = Circuit(
        [ops.H.on_each(*qreg)],
        [ops.CNOT.on(qreg[0], qreg[1])],
        [ops.T.on(qreg[2])],
        [ops.TOFFOLI.on(*qreg)],
        [ops.measure_each(*qreg)],
    )
    folded_not_squashed = fold_gates_at_random(
        circ, scale_factor=3, squash_moments=False
    )
    folded_and_squashed = fold_gates_at_random(
        circ, scale_factor=3, squash_moments=True
    )
    correct = Circuit(
        [ops.H.on_each(*qreg)] * 3,
        [ops.CNOT.on(qreg[0], qreg[1])] * 3,
        [ops.T.on(qreg[2]), ops.T.on(qreg[2]) ** -1, ops.T.on(qreg[2])],
        [ops.TOFFOLI.on(*qreg)] * 3,
        [ops.measure_each(*qreg)],
    )
    assert _equal(folded_and_squashed, folded_not_squashed)
    assert _equal(folded_and_squashed, correct)
    assert len(folded_and_squashed) == 10


def test_fold_and_squash_max_stretch():
    """Tests folding and squashing a two-qubit circuit."""
    # Test circuit:
    # 0: ───────H───────H───────H───────H───────H───
    #
    # 1: ───H───────H───────H───────H───────H───────

    # Get the test circuit
    d = 10
    qreg = LineQubit.range(2)
    circuit = Circuit()
    for i in range(d):
        circuit.insert(0, ops.H.on(qreg[i % 2]), strategy=InsertStrategy.NEW)

    folded_not_squashed = fold_gates_at_random(
        circuit, scale_factor=3.0, squash_moments=False
    )
    folded_and_squashed = fold_gates_at_random(
        circuit, scale_factor=3.0, squash_moments=True
    )
    folded_with_squash_moments_not_specified = fold_gates_at_random(
        circuit, scale_factor=3.0
    )  # Checks that the default is to squash moments

    assert len(folded_not_squashed) == 30
    assert len(folded_and_squashed) == 15
    assert len(folded_with_squash_moments_not_specified) == 15


def test_fold_and_squash_random_circuits_random_stretches():
    """Tests folding and squashing random circuits and ensures the number of
    moments in the squashed circuits is never greater than the number of
    moments in the un-squashed circuit.
    """
    rng = np.random.RandomState(seed=1)
    for trial in range(5):
        circuit = testing.random_circuit(
            qubits=8, n_moments=8, op_density=0.75
        )
        scale = 2 * rng.random() + 1
        folded_not_squashed = fold_gates_at_random(
            circuit,
            scale_factor=scale,
            squash_moments=False,
            seed=trial,
        )
        folded_and_squashed = fold_gates_at_random(
            circuit,
            scale_factor=scale,
            squash_moments=True,
            seed=trial,
        )
        assert len(folded_and_squashed) <= len(folded_not_squashed)


def test_default_weight():
    """Tests default weight of an n-qubit gates is 0.99**n."""
    qreg = LineQubit.range(3)
    assert np.isclose(_default_weight(ops.H.on(qreg[0])), 0.99)
    assert np.isclose(_default_weight(ops.CZ.on(qreg[0], qreg[1])), 0.9801)
    assert np.isclose(_default_weight(ops.TOFFOLI.on(*qreg[:3])), 0.970299)


@pytest.mark.parametrize("qiskit", [True, False])
def test_fold_local_with_fidelities(qiskit):
    qreg = LineQubit.range(3)
    circ = Circuit(
        ops.H.on_each(*qreg),
        ops.CNOT.on(qreg[0], qreg[1]),
        ops.T.on(qreg[2]),
        ops.TOFFOLI.on(*qreg),
    )
    if qiskit:
        circ = convert_from_mitiq(circ, "qiskit")
    # Only fold the Toffoli gate
    fidelities = {"H": 1.0, "T": 1.0, "CNOT": 1.0, "TOFFOLI": 0.95}
    folded = fold_gates_at_random(
        circ, scale_factor=3.0, fidelities=fidelities
    )
    correct = Circuit(
        [ops.H.on_each(*qreg)],
        [ops.CNOT.on(qreg[0], qreg[1])],
        [ops.T.on(qreg[2])],
        [ops.TOFFOLI.on(*qreg)] * 3,
    )
    if qiskit:
        folded, _ = convert_to_mitiq(folded)
        assert equal_up_to_global_phase(folded.unitary(), correct.unitary())
    else:
        assert _equal(folded, correct)


@pytest.mark.parametrize("qiskit", [True, False])
def test_fold_local_with_single_qubit_gates_fidelity_one(qiskit):
    """Tests folding only two-qubit gates by using
    fidelities = {"single": 1.}.
    """
    qreg = LineQubit.range(3)
    circ = Circuit(
        ops.H.on_each(*qreg),
        ops.CNOT.on(qreg[0], qreg[1]),
        ops.T.on(qreg[2]),
        ops.TOFFOLI.on(*qreg),
    )
    if qiskit:
        circ = convert_from_mitiq(circ, "qiskit")
    folded = fold_gates_at_random(
        circ,
        scale_factor=3.0,
        fidelities={"single": 1.0, "CNOT": 0.99, "TOFFOLI": 0.95},
    )
    correct = Circuit(
        [ops.H.on_each(*qreg)],
        [ops.CNOT.on(qreg[0], qreg[1])] * 3,
        [ops.T.on(qreg[2])],
        [ops.TOFFOLI.on(*qreg)] * 3,
    )
    if qiskit:
        assert folded.qregs == circ.qregs
        assert folded.cregs == circ.cregs
        folded, _ = convert_to_mitiq(folded)
        assert equal_up_to_global_phase(folded.unitary(), correct.unitary())
    else:
        assert _equal(folded, correct)


@pytest.mark.parametrize("qiskit", [True, False])
def test_all_gates_folded_at_max_scale_with_fidelities(qiskit):
    """Tests that all gates are folded regardless of the input fidelities when
    the scale factor is three.
    """
    rng = np.random.RandomState(1)

    qreg = LineQubit.range(3)
    circ = Circuit(
        ops.H.on_each(*qreg),
        ops.CNOT.on(qreg[0], qreg[1]),
        ops.T.on(qreg[2]),
        ops.TOFFOLI.on(*qreg),
    )
    ngates = len(list(circ.all_operations()))

    if qiskit:
        circ = convert_from_mitiq(circ, "qiskit")

    folded = fold_gates_at_random(
        circ,
        scale_factor=3.0,
        fidelities={
            "H": rng.rand(),
            "T": rng.rand(),
            "CNOT": rng.rand(),
            "TOFFOLI": rng.rand(),
        },
    )
    correct = Circuit(
        [ops.H.on_each(*qreg)] * 3,
        [ops.CNOT.on(qreg[0], qreg[1])] * 3,
        [ops.T.on(qreg[2]), ops.T.on(qreg[2]) ** -1, ops.T.on(qreg[2])],
        [ops.TOFFOLI.on(*qreg)] * 3,
    )
    if qiskit:
        assert folded.qregs == circ.qregs
        assert folded.cregs == circ.cregs
        folded, _ = convert_to_mitiq(folded)
        assert equal_up_to_global_phase(folded.unitary(), correct.unitary())
    else:
        assert _equal(folded, correct)
        assert len(list(folded.all_operations())) == 3 * ngates


def test_fold_local_raises_error_with_bad_fidelities():
    with pytest.raises(ValueError, match="Fidelities should be"):
        fold_gates_at_random(
            Circuit(), scale_factor=1.21, fidelities={"H": -1.0}
        )

    with pytest.raises(ValueError, match="Fidelities should be"):
        fold_gates_at_random(
            Circuit(), scale_factor=1.21, fidelities={"CNOT": 0.0}
        )

    with pytest.raises(ValueError, match="Fidelities should be"):
        fold_gates_at_random(
            Circuit(), scale_factor=1.21, fidelities={"triple": 1.2}
        )


@pytest.mark.parametrize("conversion_type", ("qiskit", "pyquil"))
def test_convert_from_mitiq_circuit_conversion_error(conversion_type):
    circuit = testing.random_circuit(qubits=5, n_moments=5, op_density=0.99)
    noisy = circuit.with_noise(ops.depolarize(p=0.1))

    with pytest.raises(
        CircuitConversionError, match="Circuit could not be converted from"
    ):
        convert_from_mitiq(noisy, conversion_type)


def test_convert_pyquil_to_mitiq_circuit_conversion_error():
    # Pragmas are not supported in conversions
    prog = Program(Pragma("INITIAL_REWIRING", ['"Partial"']))

    with pytest.raises(
        CircuitConversionError, match="Circuit could not be converted to"
    ):
        convert_to_mitiq(prog)


@pytest.mark.parametrize(
    "fold_method",
    (
        fold_gates_at_random,
        fold_global,
    ),
)
def test_folding_circuit_conversion_error_pyquil(fold_method):
    # Pragmas are not supported in conversions
    prog = Program(Pragma("INITIAL_REWIRING", ['"Partial"']))

    with pytest.raises(
        CircuitConversionError, match="Circuit could not be converted to"
    ):
        fold_method(prog, scale_factor=2.0)


@pytest.mark.parametrize("scale", [1, 3, 5, 9])
def test_fold_fidelity_large_scale_factor_only_twoq_gates(scale):
    qreg = LineQubit.range(2)
    circuit = Circuit(ops.H(qreg[0]), ops.CNOT(*qreg))
    folded = fold_gates_at_random(
        circuit, scale_factor=scale, fidelities={"single": 1.0}
    )
    correct = Circuit(ops.H(qreg[0]), [ops.CNOT(*qreg)] * scale)
    assert _equal(folded, correct)


def test_folding_keeps_measurement_order_with_qiskit():
    qreg, creg = QuantumRegister(2), ClassicalRegister(2)
    circuit = QuantumCircuit(qreg, creg)
    circuit.h(qreg[0])
    circuit.measure(qreg, creg)

    folded = fold_gates_at_random(circuit, scale_factor=1.0)
    assert folded == circuit


def test_create_weight_mask_with_fidelities():
    qreg = LineQubit.range(3)
    circ = Circuit(
        ops.H.on_each(*qreg),
        ops.CNOT.on(qreg[0], qreg[1]),
        ops.T.on(qreg[2]),
        ops.ISWAP.on(qreg[1], qreg[0]),
        ops.TOFFOLI.on(*qreg),
        ops.measure_each(*qreg),
    )
    # Measurement gates should be ignored
    fidelities = {
        "H": 0.9,
        "CNOT": 0.8,
        "T": 0.7,
        "TOFFOLI": 0.6,
        "ISWAP": 0.5,
    }
    weight_mask = _create_weight_mask(circ, fidelities)
    assert np.allclose(weight_mask, [0.1, 0.1, 0.1, 0.2, 0.3, 0.5, 0.4])

    fidelities = {"single": 1.0, "double": 0.5, "triple": 0.1}
    weight_mask = _create_weight_mask(circ, fidelities)
    assert np.allclose(weight_mask, [0.0, 0.0, 0.0, 0.5, 0.0, 0.5, 0.9])

    fidelities = {
        "single": 1.0,
        "double": 1.0,
        "H": 0.1,
        "CNOT": 0.2,
        "TOFFOLI": 0.3,
    }
    weight_mask = _create_weight_mask(circ, fidelities)
    assert np.allclose(weight_mask, [0.9, 0.9, 0.9, 0.8, 0.0, 0.0, 0.7])

    fidelities = {"waitgate": 1.0, "H": 0.1}
    with pytest.warns(UserWarning, match="don't currently support"):
        weight_mask = _create_weight_mask(circ, fidelities)


def test_create_fold_mask_with_real_scale_factors_at_random():
    fold_mask = _create_fold_mask(
        weight_mask=[0.1, 0.2, 0.3, 0.0],
        scale_factor=1.0,
        seed=1,
    )
    assert fold_mask == [0, 0, 0, 0]

    fold_mask = _create_fold_mask(
        weight_mask=[0.1, 0.1, 0.1, 0.1, 0.0],
        scale_factor=1.5,
        seed=2,
    )
    assert fold_mask == [0, 0, 1, 0, 0]

    fold_mask = _create_fold_mask(
        weight_mask=[1, 1, 1, 1],
        scale_factor=2,
        seed=3,
    )
    assert fold_mask == [0, 1, 0, 1]

    fold_mask = _create_fold_mask(
        weight_mask=[1, 1, 1, 1],
        scale_factor=3.9,
        seed=7,
    )
    assert fold_mask == [1, 2, 2, 1]


def test_create_fold_mask_approximates_well():
    """Check _create_fold_mask well approximates the scale factor."""
    rnd_state = np.random.RandomState(seed=0)
    for scale_factor in [1, 1.5, 1.7, 2.7, 6.7, 18.7, 19.0, 31]:
        weight_mask = [rnd_state.rand() for _ in range(100)]
        seed = rnd_state.randint(100)
        fold_mask = _create_fold_mask(
            weight_mask,
            scale_factor,
            seed=seed,
        )
        out_weights = [w + 2 * n * w for w, n in zip(weight_mask, fold_mask)]
        actual_scale = sum(out_weights) / sum(weight_mask)
        # Less than 1% error
        assert np.isclose(scale_factor / actual_scale, 1.0, atol=0.01)


def test_apply_fold_mask():
    # Test circuit
    # 0: ───H───@───@───M───
    #           │   │
    # 1: ───H───X───@───M───
    #               │
    # 2: ───H───T───X───M───
    qreg = LineQubit.range(3)
    circ = Circuit(
        [ops.H.on_each(*qreg)],
        [ops.CNOT.on(qreg[0], qreg[1])],
        [ops.T.on(qreg[2])],
        [ops.TOFFOLI.on(*qreg)],
        [ops.measure_each(*qreg)],
    )

    folded = _apply_fold_mask(circ, [0, 0, 0, 0, 0, 0])
    assert _equal(folded, circ)

    folded = _apply_fold_mask(circ, [1, 1, 1, 0, 0, 0])
    correct = Circuit([ops.H.on_each(*qreg)] * 2) + circ
    assert _equal(folded, correct)

    folded = _apply_fold_mask(circ, [0, 0, 0, 0, 0, 2])
    correct = circ[:-1] + Circuit([ops.TOFFOLI.on(*qreg)] * 4) + circ[-1]
    assert _equal(folded, correct)

    folded = _apply_fold_mask(circ, [0, 3, 0, 0, 0, 0])
    correct = Circuit([ops.H.on(qreg[1])] * 6) + circ
    assert _equal(folded, _squash_moments(correct))

    folded = _apply_fold_mask(circ, [0, 0, 1, 0, 0, 0])
    correct = Circuit([ops.H.on(qreg[2])] * 2 + circ)
    assert _equal(folded, _squash_moments(correct))

    folded = _apply_fold_mask(circ, [1, 1, 1, 1, 1, 1])
    correct = Circuit(
        [ops.H.on_each(*qreg)] * 3,
        [ops.CNOT.on(qreg[0], qreg[1])] * 3,
        [ops.T.on(qreg[2]), inverse(ops.T.on(qreg[2])), ops.T.on(qreg[2])],
        [ops.TOFFOLI.on(*qreg)] * 3,
        [ops.measure_each(*qreg)],
    )
    assert _equal(folded, correct)


def test_apply_fold_mask_wrong_size():
    qreg = LineQubit(0)
    circ = Circuit(ops.H(qreg))
    with pytest.raises(ValueError, match="have incompatible sizes"):
        _ = _apply_fold_mask(circ, [1, 1])


def test_apply_fold_mask_with_squash_moments_option():
    # Test circuit:
    # 0: ───T────────
    #
    # 1: ───T────H───
    q = LineQubit.range(2)
    circ = Circuit(
        [ops.T.on_each(*q), ops.H(q[1])],
    )
    folded = _apply_fold_mask(circ, [1, 0, 0], squash_moments=False)
    # 0: ───T───T^-1───T───────
    #
    # 1: ───T──────────────H───
    correct = Circuit(
        [ops.T.on_each(*q), inverse(ops.T(q[0])), ops.T(q[0])],
    ) + Circuit(ops.H(q[1]))
    assert _equal(folded, correct)

    # If 2 gates of the same moment are folded,
    # only 2 moments should be created and not 4.
    folded = _apply_fold_mask(circ, [1, 1, 0], squash_moments=False)
    # 0: ───T───T^-1───T───────
    #
    # 1: ───T───T^-1───T────H───
    correct = Circuit(
        [
            ops.T.on_each(*q),
            inverse(ops.T.on_each(*q)),
            ops.T.on_each(*q),
            ops.H(q[1]),
        ],
    )
    assert _equal(folded, correct)

    folded = _apply_fold_mask(circ, [1, 0, 0], squash_moments=True)
    # 0: ───T───T^-1───T───
    #
    # 1: ───T───H──────────
    correct = Circuit(
        [ops.T.on_each(*q), inverse(ops.T(q[0])), ops.T(q[0]), ops.H(q[1])],
    )
    assert _equal(folded, correct)


@pytest.mark.parametrize(
    "fold_method",
    (
        fold_gates_at_random,
        fold_global,
    ),
)
def test_scaling_with_pyquil_retains_declarations(fold_method):
    program = Program()
    theta = program.declare("theta", memory_type="REAL")
    _ = program.declare("beta", memory_type="REAL")
    program += gates.RY(theta, 0)

    scaled = fold_method(program, 1)
    assert scaled.declarations == program.declarations



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Unit tests for multivariate extrapolation inference functions."""

import numpy as np
import pytest
from cirq import Circuit, LineQubit, ops

from mitiq import SUPPORTED_PROGRAM_TYPES
from mitiq.interface import convert_from_mitiq
from mitiq.lre.inference.multivariate_richardson import (
    _full_monomial_basis_term_exponents,
    multivariate_richardson_coefficients,
    sample_matrix,
)
from mitiq.lre.multivariate_scaling.layerwise_folding import (
    multivariate_layer_scaling,
)

qreg1 = LineQubit.range(3)
test_circuit1 = Circuit(
    [ops.H.on_each(*qreg1)],
    [ops.CNOT.on(qreg1[0], qreg1[1])],
)
test_circuit2 = Circuit(
    [ops.H.on_each(*qreg1)],
    [ops.CNOT.on(qreg1[0], qreg1[1])],
    [ops.X.on(qreg1[2])],
    [ops.TOFFOLI.on(*qreg1)],
)


@pytest.mark.parametrize(
    "test_num_layers, test_degree, expected",
    [
        (1, 1, [(0,), (1,)]),
        (
            2,
            2,
            [(0, 0), (1, 0), (0, 1), (2, 0), (1, 1), (0, 2)],
        ),
        (
            3,
            2,
            [
                (0, 0, 0),
                (1, 0, 0),
                (0, 1, 0),
                (0, 0, 1),
                (2, 0, 0),
                (1, 1, 0),
                (0, 2, 0),
                (1, 0, 1),
                (0, 1, 1),
                (0, 0, 2),
            ],
        ),
    ],
)
def test_basis_exp(test_num_layers, test_degree, expected):
    assert (
        _full_monomial_basis_term_exponents(test_num_layers, test_degree)
        == expected
    )


@pytest.mark.parametrize(
    "test_num_layers, test_degree",
    [(1, 1), (2, 2), (3, 2), (10, 4)],
    # TO DO: Note need to add (100, 2) here.
    # This makes the unit test very slow.
)
def test_basis_exp_len(test_num_layers, test_degree):
    calc_dict = _full_monomial_basis_term_exponents(
        test_num_layers, test_degree
    )
    for i in calc_dict:
        assert len(i) == test_num_layers


@pytest.mark.parametrize("circuit_type", SUPPORTED_PROGRAM_TYPES.keys())
@pytest.mark.parametrize(
    "test_circ, test_degree, expected_matrix",
    [
        (
            test_circuit1,
            2,
            np.array(
                [
                    [1, 1, 1, 1, 1, 1],
                    [1, 3, 1, 9, 3, 1],
                    [1, 1, 3, 1, 3, 9],
                    [1, 5, 1, 25, 5, 1],
                    [1, 3, 3, 9, 9, 9],
                    [1, 1, 5, 1, 5, 25],
                ]
            ),
        ),
        (
            test_circuit2,
            2,
            np.array(
                [
                    [1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                    [1, 3, 1, 1, 9, 3, 1, 3, 1, 1],
                    [1, 1, 3, 1, 1, 3, 9, 1, 3, 1],
                    [1, 1, 1, 3, 1, 1, 1, 3, 3, 9],
                    [1, 5, 1, 1, 25, 5, 1, 5, 1, 1],
                    [1, 3, 3, 1, 9, 9, 9, 3, 3, 1],
                    [1, 3, 1, 3, 9, 3, 1, 9, 3, 9],
                    [1, 1, 5, 1, 1, 5, 25, 1, 5, 1],
                    [1, 1, 3, 3, 1, 3, 9, 3, 9, 9],
                    [1, 1, 1, 5, 1, 1, 1, 5, 5, 25],
                ]
            ),
        ),
    ],
)
def test_sample_matrix(test_circ, test_degree, expected_matrix, circuit_type):
    converted_circuit = convert_from_mitiq(test_circ, circuit_type)
    assert np.allclose(
        expected_matrix, sample_matrix(converted_circuit, test_degree, 1)
    )


@pytest.mark.parametrize(
    "test_circ, test_degree, test_fold_multiplier, expected_matrix",
    [
        (
            test_circuit1,
            2,
            3,
            [
                1.5555555555555578,
                -0.38888888888888934,
                -0.38888888888888934,
                0.09722222222222215,
                0.027777777777777804,
                0.09722222222222232,
            ],
        ),
        (
            test_circuit2,
            2,
            2,
            [
                2.4062499999999956,
                -0.6874999999999987,
                -0.6874999999999987,
                -0.6874999999999987,
                0.15624999999999956,
                0.06249999999999997,
                0.06249999999999997,
                0.15624999999999956,
                0.06249999999999997,
                0.15624999999999956,
            ],
        ),
    ],
)
def test_coeffs(test_circ, test_degree, test_fold_multiplier, expected_matrix):
    assert np.allclose(
        expected_matrix,
        multivariate_richardson_coefficients(
            test_circ, test_degree, test_fold_multiplier
        ),
        atol=1e-3,
    )

    assert np.isclose(
        sum(
            multivariate_richardson_coefficients(
                test_circ, test_degree, test_fold_multiplier
            )
        ),
        1.0,
    )


@pytest.mark.parametrize(
    "test_input, test_degree, test_fold_multiplier, error_msg",
    [
        (
            test_circuit1,
            0,
            1,
            "Multinomial degree must be greater than or equal to 1.",
        ),
        (
            test_circuit1,
            1,
            0,
            "Fold multiplier must be greater than or equal to 1.",
        ),
    ],
)
def test_invalid_degree_fold_multiplier_sample_matrix(
    test_input, test_degree, test_fold_multiplier, error_msg
):
    """Ensures that the args for the sample matrix
    an error for an invalid value."""
    with pytest.raises(ValueError, match=error_msg):
        sample_matrix(test_input, test_degree, test_fold_multiplier)


def test_lre_inference_with_chunking():
    """Verify the dimension of a chunked sample matrix for some input circuit
    is smaller than the non-chunked sample matrix for the same input circuit.
    """
    circ = test_circuit1 * 7
    chunked_sample_matrix_dim = sample_matrix(circ, 2, 2, num_chunks=4).shape
    non_chunked_sample_matrix_dim = sample_matrix(circ, 2, 2).shape
    assert chunked_sample_matrix_dim[0] < non_chunked_sample_matrix_dim[0]


def test_sample_matrix_numerical_stability():
    """Verify sample matrix function works for very large circuits."""
    large_circuit = Circuit([ops.H.on(LineQubit(i)) for i in range(10000)])
    matrix = sample_matrix(large_circuit, 5, 10000)
    assert np.isfinite(matrix).all()
    assert not np.isnan(matrix).any()


@pytest.mark.parametrize("num_chunks", [2, 3])
def test_eval(num_chunks):
    """Verify the number of calculated linear combination coefficients matches
    to the number of scaled chunked circuits."""
    coeffs = multivariate_richardson_coefficients(
        7 * test_circuit2, 2, 2, num_chunks
    )
    multiple_scaled_circuits = multivariate_layer_scaling(
        7 * test_circuit2, 2, 2, num_chunks
    )
    assert len(coeffs) == len(multiple_scaled_circuits)
    assert np.isclose(sum(coeffs), 1.0)



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Tests for circuit conversions."""

from typing import List

import cirq
import numpy as np
import pennylane as qml
import pytest
import qibo
import qiskit
from braket.circuits import Circuit as BKCircuit
from braket.circuits import Instruction
from braket.circuits import gates as braket_gates
from pyquil import Program, gates

from mitiq import SUPPORTED_PROGRAM_TYPES
from mitiq.interface import (
    UnsupportedCircuitError,
    accept_any_qprogram_as_input,
    accept_qprogram_and_validate,
    atomic_one_to_many_converter,
    convert_from_mitiq,
    convert_to_mitiq,
    register_mitiq_converters,
)
from mitiq.interface.mitiq_qiskit import from_qasm, to_qasm
from mitiq.utils import _equal

QASMType = str

# Cirq Bell circuit.
cirq_qreg = cirq.LineQubit.range(2)
cirq_circuit = cirq.Circuit(
    cirq.ops.H.on(cirq_qreg[0]), cirq.ops.CNOT.on(*cirq_qreg)
)

# Qiskit Bell circuit.
qiskit_qreg = qiskit.QuantumRegister(2)
qiskit_circuit = qiskit.QuantumCircuit(qiskit_qreg)
qiskit_circuit.h(qiskit_qreg[0])
qiskit_circuit.cx(*qiskit_qreg)
qasm_str = qiskit.qasm2.dumps(qiskit_circuit)


# pyQuil Bell circuit.
pyquil_circuit = Program(gates.H(0), gates.CNOT(0, 1))

# Braket Bell circuit.
braket_circuit = BKCircuit(
    [
        Instruction(braket_gates.H(), 0),
        Instruction(braket_gates.CNot(), [0, 1]),
    ]
)

circuit_types = {
    "cirq": cirq.Circuit,
    "qiskit": qiskit.QuantumCircuit,
    "pyquil": Program,
    "braket": BKCircuit,
    "pennylane": qml.tape.QuantumTape,
    "qibo": qibo.models.circuit.Circuit,
}


@accept_qprogram_and_validate
def scaling_function(circ: cirq.Circuit, *args, **kwargs) -> cirq.Circuit:
    return circ


def one_to_many_circuit_modifier(
    circ: cirq.Circuit, *args, **kwargs
) -> List[cirq.Circuit]:
    return [circ, circ[0:1] + circ]


# Apply one-to-many decorator
one_to_many_circuit_modifier = accept_qprogram_and_validate(
    one_to_many_circuit_modifier,
    one_to_many=True,
)


@accept_any_qprogram_as_input
def get_wavefunction(circ: cirq.Circuit) -> np.ndarray:
    return circ.final_state_vector()


@atomic_one_to_many_converter
def returns_several_circuits(circ: cirq.Circuit, *args, **kwargs):
    return [circ] * 5


@pytest.mark.parametrize(
    "circuit", (qiskit_circuit, pyquil_circuit, braket_circuit)
)
def test_to_mitiq(circuit):
    converted_circuit, input_type = convert_to_mitiq(circuit)
    assert _equal(converted_circuit, cirq_circuit)
    assert input_type in circuit.__module__


def test_register_from_to_mitiq():
    class CircuitStr(str):
        __module__ = "qasm"

    qasm_circuit = CircuitStr(qasm_str)

    register_mitiq_converters(
        qasm_circuit.__module__,
        convert_to_function=to_qasm,
        convert_from_function=from_qasm,
    )
    converted_circuit = convert_from_mitiq(cirq_circuit, "qasm")
    converted_qasm = CircuitStr(converted_circuit)
    circuit, input_type = convert_to_mitiq(converted_qasm)
    assert _equal(circuit, cirq_circuit)
    assert input_type == qasm_circuit.__module__


@pytest.mark.parametrize("item", ("circuit", 1, None))
def test_to_mitiq_bad_types(item):
    with pytest.raises(
        UnsupportedCircuitError,
        match="Could not determine the package of the input circuit.",
    ):
        convert_to_mitiq(item)


@pytest.mark.parametrize("to_type", SUPPORTED_PROGRAM_TYPES.keys())
def test_from_mitiq(to_type):
    converted_circuit = convert_from_mitiq(cirq_circuit, to_type)
    circuit, input_type = convert_to_mitiq(converted_circuit)
    assert _equal(circuit, cirq_circuit)
    assert input_type == to_type


def test_unsupported_circuit_error():
    class CircuitStr(str):
        __module__ = "qasm"

    mock_circuit = CircuitStr("mock")

    with pytest.raises(
        UnsupportedCircuitError,
        match="Conversion to circuit type unsupported_circuit_type",
    ):
        convert_from_mitiq(mock_circuit, "unsupported_circuit_type")


@pytest.mark.parametrize(
    "circuit_and_expected",
    [
        (cirq.Circuit(cirq.X.on(cirq.LineQubit(0))), np.array([0, 1])),
        (cirq_circuit, np.array([1, 0, 0, 1]) / np.sqrt(2)),
    ],
)
@pytest.mark.parametrize("to_type", SUPPORTED_PROGRAM_TYPES.keys())
def test_accept_any_qprogram_as_input(circuit_and_expected, to_type):
    circuit, expected = circuit_and_expected
    wavefunction = get_wavefunction(convert_from_mitiq(circuit, to_type))
    assert np.allclose(wavefunction, expected)


@pytest.mark.parametrize(
    "circuit_and_type",
    (
        (qiskit_circuit, "qiskit"),
        (pyquil_circuit, "pyquil"),
        (braket_circuit, "braket"),
    ),
)
def test_converter(circuit_and_type):
    circuit, input_type = circuit_and_type

    # Return the input type
    scaled = scaling_function(circuit)
    assert isinstance(scaled, circuit_types[input_type])

    # Return a Cirq Circuit
    cirq_scaled = scaling_function(circuit, return_mitiq=True)
    assert isinstance(cirq_scaled, cirq.Circuit)
    assert _equal(cirq_scaled, cirq_circuit)


@pytest.mark.parametrize("nbits", [1, 10])
@pytest.mark.parametrize("measure", [True, False])
def test_converter_keeps_register_structure_qiskit(nbits, measure):
    qreg = qiskit.QuantumRegister(nbits)
    creg = qiskit.ClassicalRegister(nbits)
    circ = qiskit.QuantumCircuit(qreg, creg)
    circ.h(qreg)

    if measure:
        circ.measure(qreg, creg)

    scaled = scaling_function(circ)

    assert scaled.qregs == circ.qregs
    assert scaled.cregs == circ.cregs
    assert scaled == circ


@pytest.mark.parametrize("nbits", [1, 10])
@pytest.mark.parametrize("measure", [True, False])
def test_converter_keeps_register_structure_qiskit_one_to_many(nbits, measure):
    qreg = qiskit.QuantumRegister(nbits)
    creg = qiskit.ClassicalRegister(nbits)
    circ = qiskit.QuantumCircuit(qreg, creg)
    circ.h(qreg)

    if measure:
        circ.measure(qreg, creg)

    out_circuits = one_to_many_circuit_modifier(circ)

    for out_circ in out_circuits:
        assert out_circ.qregs == circ.qregs
        assert out_circ.cregs == circ.cregs
        print(out_circ)
    assert out_circuits[0] == circ
    assert out_circuits[1] != circ


@pytest.mark.parametrize("to_type", SUPPORTED_PROGRAM_TYPES.keys())
def test_atomic_one_to_many_converter(to_type):
    circuit = convert_from_mitiq(cirq_circuit, to_type)
    circuits = returns_several_circuits(circuit)
    for circuit in circuits:
        assert isinstance(circuit, circuit_types[to_type])

    circuits = returns_several_circuits(circuit, return_mitiq=True)
    for circuit in circuits:
        assert isinstance(circuit, cirq.Circuit)


def test_noise_scaling_converter_with_qiskit_idle_qubits_and_barriers():
    """Idle qubits must be preserved even if the input has barriers.
    Test input:
         ┌───┐ ░
    q_0: ┤ X ├─░─
         └───┘ ░
    q_1: ──────░─
         ┌───┐ ░
    q_2: ┤ X ├─░─
         └───┘ ░
    q_3: ────────
    Expected output:
         ┌───┐
    q_0: ┤ X ├
         └───┘
    q_1: ─────
         ┌───┐
    q_2: ┤ X ├
         └───┘
    q_3: ─────
    """
    test_circuit_qiskit = qiskit.QuantumCircuit(4)
    test_circuit_qiskit.x(0)
    test_circuit_qiskit.x(2)
    test_circuit_qiskit.barrier(0, 1, 2)
    test_copy = test_circuit_qiskit.copy()

    scaled = scaling_function(test_circuit_qiskit)
    # Mitiq is expected to remove qiskit barriers
    expected = qiskit.QuantumCircuit(4)
    expected.x(0)
    expected.x(2)
    assert scaled == expected
    # Mitiq should not mutate the input circuit
    assert test_circuit_qiskit == test_copy



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Unit tests for high-level DDD tools."""

from typing import List

import cirq
import numpy as np
from pytest import mark

from mitiq import QPROGRAM, SUPPORTED_PROGRAM_TYPES, Executor
from mitiq.ddd import (
    construct_circuits,
    ddd_decorator,
    execute_with_ddd,
    mitigate_executor,
)
from mitiq.ddd.rules import xx, xyxy, yy
from mitiq.interface import convert_from_mitiq, convert_to_mitiq
from mitiq.interface.mitiq_cirq import compute_density_matrix
from mitiq.pec.tests.test_pec import (
    batched_executor,
    noiseless_serial_executor,
    serial_executor,
)

# A layer of X gates is useful otherwise amplitude damping is not effective
x_layer = cirq.Circuit(cirq.X.on_each(cirq.LineQubit.range(7)))
circuit_cirq_a = x_layer + cirq.Circuit(
    cirq.SWAP(q, q + 1) for q in cirq.LineQubit.range(7)
)
# Manually append inverse to avoid conversions of SWAP^-1.
circuit_cirq_a += (
    cirq.Circuit(cirq.SWAP(q, q + 1) for q in cirq.LineQubit.range(7)[::-1])
    + x_layer
)

circuit_cirq_b = x_layer[:4] + cirq.Circuit(
    cirq.CNOT(q, q + 1) for q in cirq.LineQubit.range(4)
)
circuit_cirq_b += cirq.inverse(circuit_cirq_b)


def amp_damp_executor(circuit: QPROGRAM, noise: float = 0.005) -> float:
    circuit, _ = convert_to_mitiq(circuit)
    return compute_density_matrix(
        circuit, noise_model_function=cirq.amplitude_damp, noise_level=(noise,)
    )[0, 0].real


@mark.parametrize("circuit_type", SUPPORTED_PROGRAM_TYPES.keys())
@mark.parametrize("circuit", [circuit_cirq_a, circuit_cirq_b])
@mark.parametrize("rule", [xx, yy, xyxy])
def test_execute_with_ddd_without_noise(circuit_type, circuit, rule):
    """Tests that execute_with_ddd preserves expected results
    in the absence of noise.
    """
    circuit = convert_from_mitiq(circuit, circuit_type)
    true_noiseless_value = 1.0
    unmitigated = noiseless_serial_executor(circuit)
    mitigated = execute_with_ddd(
        circuit,
        executor=noiseless_serial_executor,
        rule=rule,
    )
    error_unmitigated = abs(unmitigated - true_noiseless_value)
    error_mitigated = abs(mitigated - true_noiseless_value)
    assert np.isclose(error_unmitigated, error_mitigated)


@mark.parametrize("circuit_type", SUPPORTED_PROGRAM_TYPES.keys())
@mark.parametrize("circuit", [circuit_cirq_a, circuit_cirq_b])
@mark.parametrize("executor", [serial_executor, batched_executor])
@mark.parametrize("rule", [xx, yy, xyxy])
def test_execute_with_ddd_and_depolarizing_noise(
    circuit_type, circuit, executor, rule
):
    """Tests that with execute_with_ddd the error of a noisy
    expectation value is unchanged with depolarizing noise.
    """
    circuit = convert_from_mitiq(circuit, circuit_type)
    true_noiseless_value = 1.0
    unmitigated = serial_executor(circuit)
    mitigated = execute_with_ddd(
        circuit,
        executor,
        rule=rule,
    )
    error_unmitigated = abs(unmitigated - true_noiseless_value)
    error_mitigated = abs(mitigated - true_noiseless_value)

    # For moment-based depolarizing noise DDD should
    # have no effect (since noise commutes with DDD gates).
    assert np.isclose(error_mitigated, error_unmitigated)


@mark.parametrize("circuit_type", SUPPORTED_PROGRAM_TYPES.keys())
@mark.parametrize("rule", [xx, yy, xyxy])
def test_execute_with_ddd_and_damping_noise(circuit_type, rule):
    """Tests that with execute_with_ddd the error of a noisy
    expectation value is unchanged with depolarizing noise.
    """
    circuit = convert_from_mitiq(circuit_cirq_a, circuit_type)
    true_noiseless_value = 1.0
    unmitigated = amp_damp_executor(circuit)
    mitigated = execute_with_ddd(
        circuit,
        amp_damp_executor,
        rule=rule,
    )
    error_unmitigated = abs(unmitigated - true_noiseless_value)
    error_mitigated = abs(mitigated - true_noiseless_value)

    assert error_mitigated < error_unmitigated


@mark.parametrize("executor", [serial_executor, batched_executor])
def test_execute_with_ddd_with_num_trials(executor):
    """Tests the option num_trials of execute_with_ddd."""
    executor = Executor(executor)
    mitigated_1 = execute_with_ddd(
        circuit_cirq_a,
        executor,
        rule=xx,
        num_trials=1,
    )
    assert executor.calls_to_executor == 1
    assert len(executor.executed_circuits) == 1

    mitigated_2 = execute_with_ddd(
        circuit_cirq_a,
        executor,
        rule=xx,
        num_trials=2,
    )
    # Note executor contains the history of both experiments
    if executor.can_batch:
        assert executor.calls_to_executor == 2
    else:
        assert executor.calls_to_executor == 3
    assert len(executor.executed_circuits) == 3

    # For deterministic DDD sequences num_trials is irrelevant
    assert np.isclose(mitigated_1, mitigated_2)


def test_execute_with_ddd_with_full_output():
    """Tests the option full_output of execute_with_ddd."""
    executor = Executor(noiseless_serial_executor)

    ddd_value, ddd_data = execute_with_ddd(
        circuit_cirq_a,
        executor,
        rule=xx,
        num_trials=2,
        full_output=True,
    )
    assert len(executor.executed_circuits) == 2
    assert len(ddd_data["circuits_with_ddd"]) == 2
    assert len(ddd_data["ddd_trials"]) == 2
    assert ddd_data["ddd_value"] == ddd_value
    # For a deterministic rule
    assert ddd_data["ddd_trials"][0] == ddd_data["ddd_trials"][1]


def test_mitigate_executor_ddd():
    ddd_value = execute_with_ddd(
        circuit_cirq_a,
        serial_executor,
        rule=xx,
    )
    mitigated_executor = mitigate_executor(serial_executor, rule=xx)
    assert np.isclose(mitigated_executor(circuit_cirq_a), ddd_value)

    batched_mitigated_executor = mitigate_executor(batched_executor, rule=xx)
    assert np.isclose(
        *batched_mitigated_executor([circuit_cirq_a] * 3), ddd_value
    )


def test_ddd_decorator():
    ddd_value = execute_with_ddd(
        circuit_cirq_a,
        serial_executor,
        rule=xx,
    )

    @ddd_decorator(rule=xx)
    def my_serial_executor(circuit):
        return serial_executor(circuit)

    assert np.isclose(my_serial_executor(circuit_cirq_a), ddd_value)

    # Test batched executors too
    @ddd_decorator(rule=xx)
    def my_batched_executor(circuits) -> List[float]:
        return batched_executor(circuits)

    assert np.isclose(*my_batched_executor([circuit_cirq_a]), ddd_value)


def test_ddd_decorator_with_rule_args():
    """Tests that rule_args option is working."""
    unmitigated = amp_damp_executor(circuit_cirq_a)

    @ddd_decorator(rule=xx)
    def exec_xx(circuit):
        return amp_damp_executor(circuit)

    mitigated = exec_xx(circuit_cirq_a)
    assert unmitigated < mitigated

    @ddd_decorator(rule=xx, rule_args={"spacing": 100})
    def exec_xx_large_spacing(circuit):
        return amp_damp_executor(circuit)

    mitigated_large_spacing = exec_xx_large_spacing(circuit_cirq_a)
    # With very large spacing DDD sequences should not fit in the circuit.
    # So we should get the same result as without mitigation.
    assert np.isclose(unmitigated, mitigated_large_spacing)

    @ddd_decorator(rule=xx, rule_args={"spacing": 1})
    def exec_xx_small_spacing(circuit):
        return amp_damp_executor(circuit)

    mitigated_small_spacing = exec_xx_small_spacing(circuit_cirq_a)
    # With small spacing results can be better or worst than default spacing.
    # What is important to test is getting different results.
    assert not np.isclose(unmitigated, mitigated_small_spacing)
    assert not np.isclose(mitigated_large_spacing, mitigated_small_spacing)


@mark.parametrize("num_trials", [1, 10, 20, 30])
def test_num_trials_generates_circuits(num_trials: int):
    """Test that the number of generated circuits follows num_trials."""

    circuits = construct_circuits(
        circuit_cirq_a, rule=xx, num_trials=num_trials
    )

    assert num_trials == len(circuits)



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Unit tests for parameter scaling."""

from copy import deepcopy

import numpy as np
import pytest
from cirq import CSWAP, Circuit, LineQubit, ZPowGate, ops

from mitiq.utils import _equal
from mitiq.zne.scaling.parameter import (
    CircuitMismatchException,
    GateTypeException,
    _generate_parameter_calibration_circuit,
    _get_base_gate,
    compute_parameter_variance,
    scale_parameters,
)


def test_identity_scale_1q():
    """Tests that when scale factor = 1, the circuit is the
    same.
    """
    qreg = LineQubit.range(3)
    circ = Circuit([ops.X.on_each(qreg)], [ops.Y.on(qreg[0])])
    scaled = scale_parameters(circ, scale_factor=1, base_variance=0.001)
    assert _equal(circ, scaled)


def test_non_identity_scale_1q():
    """Tests that when scale factor = 1, the circuit is the
    same.
    """
    qreg = LineQubit.range(3)
    circ = Circuit(
        [ops.rx(np.pi * 1.0).on_each(qreg)], [ops.ry(np.pi * 1.0).on(qreg[0])]
    )
    np.random.seed(42)
    stretch = 2
    base_noise = 0.001
    noises = np.random.normal(
        loc=0.0, scale=np.sqrt((stretch - 1) * base_noise), size=(4,)
    )
    np.random.seed(42)

    scaled = scale_parameters(
        circ, scale_factor=stretch, base_variance=base_noise, seed=42
    )
    result = []
    for moment in scaled:
        for op in moment.operations:
            gate = deepcopy(op.gate)
            param = gate.exponent
            result.append(param * np.pi - np.pi)
    assert np.all(np.isclose(result - noises, 0))


def test_identity_scale_2q():
    """Tests that when scale factor = 1, the circuit is the
    same.
    """
    qreg = LineQubit.range(2)
    circ = Circuit([ops.CNOT.on(qreg[0], qreg[1])])
    scaled = scale_parameters(circ, scale_factor=1, base_variance=0.001)
    assert _equal(circ, scaled)


def test_non_identity_scale_2q():
    """Tests that when scale factor = 1, the circuit is the
    same.
    """
    qreg = LineQubit.range(2)
    circ = Circuit([ops.CNOT.on(qreg[0], qreg[1])])
    np.random.seed(42)
    stretch = 2
    base_noise = 0.001
    noises = np.random.normal(
        loc=0.0, scale=np.sqrt((stretch - 1) * base_noise), size=(1,)
    )
    np.random.seed(42)
    scaled = scale_parameters(
        circ, scale_factor=stretch, base_variance=base_noise, seed=42
    )
    result = []
    for moment in scaled:
        for op in moment.operations:
            gate = deepcopy(op.gate)
            param = gate.exponent
            result.append(param * np.pi - np.pi)
    assert np.all(np.isclose(result - noises, 0))


def test_scale_with_measurement():
    """Tests that we ignore measurement gates.

    Test circuit:
    0: ───H───T───@───M───
                  │   │
    1: ───H───M───┼───┼───
                  │   │
    2: ───H───────X───M───

    """
    qreg = LineQubit.range(3)
    circ = Circuit(
        [ops.H.on_each(qreg)],
        [ops.T.on(qreg[0])],
        [ops.measure(qreg[1])],
        [ops.CNOT.on(qreg[0], qreg[2])],
        [ops.measure(qreg[0], qreg[2])],
    )
    scaled = scale_parameters(circ, scale_factor=1, base_variance=0.001)
    assert _equal(circ, scaled)


def test_gate_type():
    qreg = LineQubit.range(3)
    allowed_op = ops.H.on(qreg[0])
    _get_base_gate(allowed_op.gate)
    with pytest.raises(GateTypeException):
        forbidden_op = CSWAP(qreg[0], qreg[1], qreg[2])
        _get_base_gate(forbidden_op)


def test_compute_parameter_variance():
    def noiseless_executor_mock(circuit):
        return 1

    # Perfect executor should have sigma = 0
    qubit = LineQubit(0)
    gate = ops.H.on(qubit).gate
    sigma = compute_parameter_variance(
        noiseless_executor_mock, gate, qubit, depth=10
    )
    assert sigma == 0

    # Perfectly imperfect executor should have sigma = inf
    def noisy_executor_mock(circuit):
        return 0.5

    qubit = LineQubit(0)
    gate = ops.H.on(qubit).gate
    with pytest.warns(RuntimeWarning):
        # Runtime warning for divide by zero
        sigma = compute_parameter_variance(
            noisy_executor_mock, gate, qubit, depth=10
        )
    assert sigma == np.inf


def test_generate_parameter_calibration_circuit():
    """Tests generating a simple Parameter Calibration circuit"""
    n_qubits = 1
    qubits = LineQubit.range(n_qubits)
    depth = 10
    circuit = _generate_parameter_calibration_circuit(qubits, depth, ZPowGate)
    assert len(circuit) == depth
    # Make sure the exponents in the
    for i in range(len(circuit)):
        assert circuit[i].operations[0].gate.exponent == 2 * np.pi / depth


def test_generate_parameter_calibration_circuit_failure():
    """Tests that parameter calibration circuit generation fails because there
    are too many qubits"""
    n_qubits = 3
    qubits = LineQubit.range(n_qubits)
    depth = 10
    # Should raise exception because too many qubits
    with pytest.raises(CircuitMismatchException):
        _generate_parameter_calibration_circuit(qubits, depth, ZPowGate)



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

from itertools import product

import numpy as np
from cirq import (
    CNOT,
    CZ,
    AmplitudeDampingChannel,
    Circuit,
    DepolarizingChannel,
    H,
    I,
    LineQubit,
    ResetChannel,
    X,
    Y,
    Z,
    kraus,
    reset,
)
from pytest import mark, raises

from mitiq.interface import convert_from_mitiq
from mitiq.pec.channels import (
    _circuit_to_choi,
    _operation_to_choi,
    choi_to_super,
    kraus_to_choi,
    kraus_to_super,
)
from mitiq.pec.representations import (
    _represent_operation_with_amplitude_damping_noise,
    amplitude_damping_kraus,
    global_depolarizing_kraus,
    represent_operation_with_local_depolarizing_noise,
)
from mitiq.pec.representations.optimal import (
    find_optimal_representation,
    minimize_one_norm,
)
from mitiq.pec.types import NoisyOperation


def test_minimize_one_norm_failure_error():
    ideal_matrix = np.random.rand(2, 2)
    basis_matrices = [np.random.rand(2, 2)]
    with raises(RuntimeError, match="optimal representation failed"):
        minimize_one_norm(ideal_matrix, basis_matrices)


def test_minimize_one_norm_with_depolarized_choi():
    for noise_level in [0.01, 0.02, 0.03]:
        q = LineQubit(0)
        ideal_matrix = _operation_to_choi(H(q))
        basis_matrices = [
            _operation_to_choi(
                [H(q), gate(q), DepolarizingChannel(noise_level, 1)(q)]
            )
            for gate in [I, X, Y, Z, H]
        ]
        optimal_coeffs = minimize_one_norm(ideal_matrix, basis_matrices)
        represented_mat = sum(
            [eta * mat for eta, mat in zip(optimal_coeffs, basis_matrices)]
        )
        assert np.allclose(ideal_matrix, represented_mat)

        # Optimal analytic result by Takagi (arXiv:2006.12509)
        eps = 4.0 / 3.0 * noise_level
        expected = (1.0 + 0.5 * eps) / (1.0 - eps)
        assert np.isclose(np.linalg.norm(optimal_coeffs, 1), expected)


def test_minimize_one_norm_with_depolarized_superoperators():
    for noise_level in [0.01, 0.02, 0.03]:
        depo_kraus = global_depolarizing_kraus(noise_level, num_qubits=1)
        depo_super = kraus_to_super(depo_kraus)
        ideal_matrix = kraus_to_super(kraus(H))
        basis_matrices = [
            depo_super @ kraus_to_super(kraus(gate)) @ ideal_matrix
            for gate in [I, X, Y, Z, H]
        ]
        optimal_coeffs = minimize_one_norm(ideal_matrix, basis_matrices)
        represented_mat = sum(
            [eta * mat for eta, mat in zip(optimal_coeffs, basis_matrices)]
        )
        assert np.allclose(ideal_matrix, represented_mat)

        # Optimal analytic result by Takagi (arXiv:2006.12509)
        eps = 4.0 / 3.0 * noise_level
        expected = (1.0 + 0.5 * eps) / (1.0 - eps)
        assert np.isclose(np.linalg.norm(optimal_coeffs, 1), expected)


def test_minimize_one_norm_with_amp_damp_choi():
    for noise_level in [0.01, 0.02, 0.03]:
        q = LineQubit(0)
        ideal_matrix = _operation_to_choi(H(q))
        basis_matrices = [
            _operation_to_choi(
                [H(q), gate(q), AmplitudeDampingChannel(noise_level)(q)]
            )
            for gate in [I, Z]
        ]
        # Append reset channel
        reset_kraus = kraus(ResetChannel())
        basis_matrices.append(kraus_to_choi(reset_kraus))
        optimal_coeffs = minimize_one_norm(ideal_matrix, basis_matrices)
        represented_mat = sum(
            [eta * mat for eta, mat in zip(optimal_coeffs, basis_matrices)]
        )
        assert np.allclose(ideal_matrix, represented_mat)

        # Optimal analytic result by Takagi (arXiv:2006.12509)
        expected = (1.0 + noise_level) / (1.0 - noise_level)
        assert np.isclose(np.linalg.norm(optimal_coeffs, 1), expected)


def test_minimize_one_norm_with_amp_damp_superoperators():
    for noise_level in [0.01, 0.02, 0.03]:
        damp_kraus = amplitude_damping_kraus(noise_level, num_qubits=1)
        damp_super = kraus_to_super(damp_kraus)
        ideal_matrix = kraus_to_super(kraus(H))
        basis_matrices = [
            damp_super @ kraus_to_super(kraus(gate)) @ ideal_matrix
            for gate in [I, Z]
        ]
        # Append reset channel
        reset_kraus = kraus(ResetChannel())
        basis_matrices.append(kraus_to_super(reset_kraus))
        optimal_coeffs = minimize_one_norm(
            ideal_matrix, basis_matrices, tol=1.0e-6
        )
        represented_mat = sum(
            [eta * mat for eta, mat in zip(optimal_coeffs, basis_matrices)]
        )
        assert np.allclose(ideal_matrix, represented_mat)

        # Optimal analytic result by Takagi (arXiv:2006.12509)
        expected = (1.0 + noise_level) / (1.0 - noise_level)
        assert np.isclose(np.linalg.norm(optimal_coeffs, 1), expected)


def test_minimize_one_norm_tolerance():
    depo_kraus = global_depolarizing_kraus(noise_level=0.1, num_qubits=1)
    depo_super = kraus_to_super(depo_kraus)
    ideal_matrix = kraus_to_super(kraus(H))
    basis_matrices = [
        depo_super @ kraus_to_super(kraus(gate)) @ ideal_matrix
        for gate in [I, X, Y, Z]
    ]
    previous_minimum = 0.0
    previous_error = 1.0
    for tol in [1.0e-2, 1.0e-4, 1.0e-6, 1.0e-8]:
        optimal_coeffs = minimize_one_norm(ideal_matrix, basis_matrices, tol)
        represented_mat = sum(
            [eta * mat for eta, mat in zip(optimal_coeffs, basis_matrices)]
        )
        worst_case_error = np.max(abs(ideal_matrix - represented_mat))
        minimum = np.linalg.norm(optimal_coeffs, 1)
        # Reducing "tol" should decrease the worst case error
        # and should also increase the objective function
        assert worst_case_error < previous_error
        assert minimum > previous_minimum
        previous_error = worst_case_error
        previous_minimum = minimum


@mark.parametrize("circ_type", ["cirq", "qiskit", "pyquil", "braket"])
def test_find_optimal_representation_depolarizing_two_qubit_gates(circ_type):
    """Test optimal representation agrees with a known analytic result."""
    for ideal_gate, noise_level in product([CNOT, CZ], [0.1, 0.5]):
        q = LineQubit.range(2)
        ideal_op = Circuit(ideal_gate(*q))
        implementable_circuits = [Circuit(ideal_op)]
        # Append two-qubit-gate with Paulis on one qubit
        for gate in [X, Y, Z]:
            implementable_circuits.append(Circuit([ideal_op, gate(q[0])]))
            implementable_circuits.append(Circuit([ideal_op, gate(q[1])]))
        # Append two-qubit gate with Paulis on both qubits
        for gate_a, gate_b in product([X, Y, Z], repeat=2):
            implementable_circuits.append(
                Circuit([ideal_op, gate_a(q[0]), gate_b(q[1])])
            )
        noisy_circuits = [
            circ + Circuit(DepolarizingChannel(noise_level).on_each(*q))
            for circ in implementable_circuits
        ]
        super_operators = [
            choi_to_super(_circuit_to_choi(circ)) for circ in noisy_circuits
        ]

        # Define circuits with native types
        implementable_native = [
            convert_from_mitiq(c, circ_type) for c in implementable_circuits
        ]
        ideal_op_native = convert_from_mitiq(ideal_op, circ_type)

        noisy_operations = [
            NoisyOperation(ideal, real)
            for ideal, real in zip(implementable_native, super_operators)
        ]

        # Find optimal representation
        rep = find_optimal_representation(
            ideal_op_native, noisy_operations, tol=1.0e-8
        )
        # Expected analytical result
        expected_rep = represent_operation_with_local_depolarizing_noise(
            ideal_op_native,
            noise_level,
        )
        assert np.allclose(np.sort(rep.coeffs), np.sort(expected_rep.coeffs))
        assert rep == expected_rep


@mark.parametrize("circ_type", ["cirq", "qiskit", "pyquil", "braket"])
def test_find_optimal_representation_single_qubit_depolarizing(circ_type):
    """Test optimal representation agrees with a known analytic result."""
    for ideal_gate, noise_level in product([X, Y, H], [0.1, 0.3]):
        q = LineQubit(0)

        ideal_op = Circuit(ideal_gate(q))
        implementable_circuits = [Circuit(ideal_op)]
        # Add ideal gate followed by Paulis
        for gate in [X, Y, Z]:
            implementable_circuits.append(Circuit([ideal_op, gate(q)]))

        noisy_circuits = [
            circ + Circuit(DepolarizingChannel(noise_level).on_each(q))
            for circ in implementable_circuits
        ]
        super_operators = [
            choi_to_super(_circuit_to_choi(circ)) for circ in noisy_circuits
        ]

        # Define circuits with native types
        implementable_native = [
            convert_from_mitiq(c, circ_type) for c in implementable_circuits
        ]
        ideal_op_native = convert_from_mitiq(ideal_op, circ_type)

        noisy_operations = [
            NoisyOperation(ideal, real)
            for ideal, real in zip(implementable_native, super_operators)
        ]
        # Find optimal representation
        rep = find_optimal_representation(
            ideal_op_native,
            noisy_operations,
            tol=1.0e-8,
        )
        # Expected analytical result
        expected_rep = represent_operation_with_local_depolarizing_noise(
            ideal_op_native,
            noise_level,
        )
        assert np.allclose(np.sort(rep.coeffs), np.sort(expected_rep.coeffs))
        assert rep == expected_rep


# After fixing the GitHub issue gh-702, other circuit types could be added.
@mark.parametrize("circ_type", ["cirq"])
def test_find_optimal_representation_single_qubit_amp_damping(circ_type):
    """Test optimal representation of agrees with a known analytic result."""
    for ideal_gate, noise_level in product([X, Y, H], [0.1, 0.3]):
        q = LineQubit(0)

        ideal_op = Circuit(ideal_gate(q))
        implementable_circuits = [Circuit(ideal_op)]
        # Add ideal gate followed by Paulis and reset
        for gate in [Z, reset]:
            implementable_circuits.append(Circuit([ideal_op, gate(q)]))

        noisy_circuits = [
            circ + Circuit(AmplitudeDampingChannel(noise_level).on_each(q))
            for circ in implementable_circuits
        ]

        super_operators = [
            choi_to_super(_circuit_to_choi(circ)) for circ in noisy_circuits
        ]

        # Define circuits with native types
        implementable_native = [
            convert_from_mitiq(c, circ_type) for c in implementable_circuits
        ]
        ideal_op_native = convert_from_mitiq(ideal_op, circ_type)

        noisy_operations = [
            NoisyOperation(ideal, real)
            for ideal, real in zip(implementable_native, super_operators)
        ]
        # Find optimal representation
        rep = find_optimal_representation(
            ideal_op_native,
            noisy_operations,
            tol=1.0e-7,
            initial_guess=[0, 0, 0],
        )
        # Expected analytical result
        expected_rep = _represent_operation_with_amplitude_damping_noise(
            ideal_op_native,
            noise_level,
        )
        assert np.allclose(np.sort(rep.coeffs), np.sort(expected_rep.coeffs))
        assert rep == expected_rep


def test_find_optimal_representation_no_superoperator_error():
    q = LineQubit(0)
    # Define noisy operation without superoperator matrix
    noisy_op = NoisyOperation(Circuit(X(q)))
    with raises(ValueError, match="numerical superoperator matrix"):
        find_optimal_representation(Circuit(X(q)), [noisy_op])


def test_initial_guess_in_minimize_one_norm():
    for noise_level in [0.7, 0.9]:
        depo_kraus = global_depolarizing_kraus(noise_level, num_qubits=1)
        depo_super = kraus_to_super(depo_kraus)
        ideal_matrix = kraus_to_super(kraus(H))
        basis_matrices = [
            depo_super @ kraus_to_super(kraus(gate)) @ ideal_matrix
            for gate in [I, X, Y, Z, H]
        ]
        optimal_coeffs = minimize_one_norm(
            ideal_matrix,
            basis_matrices,
            initial_guess=[1.0, 1.0, 1.0, 1.0, 1.0],
        )
        represented_mat = sum(
            [eta * mat for eta, mat in zip(optimal_coeffs, basis_matrices)]
        )
        assert np.allclose(ideal_matrix, represented_mat)

        # Test bad argument
        with raises(ValueError, match="shapes"):
            minimize_one_norm(
                ideal_matrix,
                basis_matrices,
                initial_guess=[1],
            )



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.


"""Tests for QPE benchmarking circuits."""

import cirq
import pytest

from mitiq import SUPPORTED_PROGRAM_TYPES
from mitiq.benchmarks.qpe_circuits import generate_qpe_circuit
from mitiq.utils import _equal


def test_bad_qubit_number():
    for n in (-1, 0):
        with pytest.raises(
            ValueError,
            match="{} is invalid for the number of eigenvalue reg qubits.",
        ):
            generate_qpe_circuit(n, cirq.T)


def test_bad_gate():
    for g in (cirq.CNOT, cirq.SWAP):
        with pytest.raises(
            ValueError,
            match="This QPE method only works for 1-qubit gates.",
        ):
            generate_qpe_circuit(4, g)


def test_bad_evalue_register():
    for g in (cirq.X, cirq.T):
        with pytest.raises(
            ValueError,
            match="The eigenvalue reg must be larger than the eigenstate reg.",
        ):
            generate_qpe_circuit(1, g)


@pytest.mark.parametrize("return_type", SUPPORTED_PROGRAM_TYPES.keys())
def test_conversion(return_type):
    circuit = generate_qpe_circuit(3, cirq.Y, return_type)
    assert return_type in circuit.__module__


def test_circuit_PauliX():
    generated_circuit = generate_qpe_circuit(3, cirq.X)
    qubits = cirq.LineQubit.range(3 + 1)
    expected_circuit = cirq.Circuit(
        [
            cirq.Moment(
                cirq.H(qubits[0]),
                cirq.H(qubits[1]),
                cirq.H(qubits[2]),
            ),
            cirq.Moment(
                cirq.CNOT(qubits[2], qubits[3]),
            ),
            cirq.Moment(
                cirq.CNOT(qubits[1], qubits[3]),
            ),
            cirq.Moment(
                cirq.CNOT(qubits[1], qubits[3]),
            ),
            cirq.Moment(
                cirq.CNOT(qubits[0], qubits[3]),
            ),
            cirq.Moment(
                cirq.CNOT(qubits[0], qubits[3]),
            ),
            cirq.Moment(
                cirq.CNOT(qubits[0], qubits[3]),
            ),
            cirq.Moment(
                cirq.CNOT(qubits[0], qubits[3]),
            ),
            cirq.Moment(
                cirq.SWAP(qubits[0], qubits[2]),
            ),
            cirq.Moment(
                cirq.H(qubits[0]),
            ),
            cirq.Moment(
                (cirq.CNOT**-1.0).on(qubits[0], qubits[1]),
            ),
            cirq.Moment(
                cirq.H(qubits[1]),
            ),
            cirq.Moment(
                (cirq.CNOT**-1.0).on(qubits[0], qubits[2]),
            ),
            cirq.Moment(
                (cirq.CNOT**-1.0).on(qubits[1], qubits[2]),
            ),
            cirq.Moment(
                cirq.H(qubits[2]),
            ),
        ]
    )
    assert _equal(generated_circuit, expected_circuit)


def test_circuit_QPE_TGate():
    generated_circuit = generate_qpe_circuit(3, cirq.T)
    qubits = cirq.LineQubit.range(3 + 1)
    expected_circuit = cirq.Circuit(
        [
            cirq.Moment(
                cirq.H(qubits[0]),
                cirq.H(qubits[1]),
                cirq.H(qubits[2]),
            ),
            cirq.Moment(
                (cirq.CZ**0.25).on(qubits[2], qubits[3]),
            ),
            cirq.Moment(
                (cirq.CZ**0.25).on(qubits[1], qubits[3]),
            ),
            cirq.Moment(
                (cirq.CZ**0.25).on(qubits[1], qubits[3]),
            ),
            cirq.Moment(
                (cirq.CZ**0.25).on(qubits[0], qubits[3]),
            ),
            cirq.Moment(
                (cirq.CZ**0.25).on(qubits[0], qubits[3]),
            ),
            cirq.Moment(
                (cirq.CZ**0.25).on(qubits[0], qubits[3]),
            ),
            cirq.Moment(
                (cirq.CZ**0.25).on(qubits[0], qubits[3]),
            ),
            cirq.Moment(
                cirq.SWAP(qubits[0], qubits[2]),
            ),
            cirq.Moment(
                cirq.H(qubits[0]),
            ),
            cirq.Moment(
                (cirq.CZ**-0.25).on(qubits[0], qubits[1]),
            ),
            cirq.Moment(
                cirq.H(qubits[1]),
            ),
            cirq.Moment(
                (cirq.CZ**-0.25).on(qubits[0], qubits[2]),
            ),
            cirq.Moment(
                (cirq.CZ**-0.25).on(qubits[1], qubits[2]),
            ),
            cirq.Moment(
                cirq.H(qubits[2]),
            ),
        ]
    )
    assert _equal(generated_circuit, expected_circuit)

    generated_circuit_check_default_option = generate_qpe_circuit(3)
    assert _equal(generated_circuit_check_default_option, expected_circuit)


def test_phase_angle_estimation():
    def _binary_to_int(input_array):
        result_string = ""
        for i in range(len(input_array)):
            result_string += str(input_array[i])
        return int(result_string, 2)

    n_eigstate_reg = 3
    # initialize the eigenstate reg to |+> by using H
    # + add measurements on ereg register
    # expect eigenvalue to be +1
    # the constant calculated is n in exp(2 pi i n)
    qubits = cirq.LineQubit.range(n_eigstate_reg + 1)
    PauliX_on_0_qft_circuit = (
        cirq.Circuit(cirq.H(qubits[3]))
        + generate_qpe_circuit(3, cirq.X)
        + cirq.measure(qubits[0:3])
    )
    simulator = cirq.Simulator()
    result1 = simulator.run(PauliX_on_0_qft_circuit)
    result_array1 = result1.records["q(0),q(1),q(2)"]
    result1 = _binary_to_int(result_array1[0][0])
    calculated_constant = result1 / 2**n_eigstate_reg
    assert calculated_constant == 0

    # initialize the eigenstate reg to |-> by using X and H after Pauli X Gate
    # + add measurements on ereg register
    # expect eigenvalue to be -1
    # the constant calculated is n in exp(2 pi i n)
    qubits = cirq.LineQubit.range(n_eigstate_reg + 1)
    PauliX_on_1_qft_circuit = (
        cirq.Circuit(cirq.X(qubits[3]), cirq.H(qubits[3]))
        + generate_qpe_circuit(3, cirq.X)
        + cirq.measure(qubits[0:3])
    )
    simulator = cirq.Simulator()
    result2 = simulator.run(PauliX_on_1_qft_circuit)
    result_array2 = result2.records["q(0),q(1),q(2)"]
    result2 = _binary_to_int(result_array2[0][0])
    calculated_constant = result2 / 2**n_eigstate_reg
    assert calculated_constant == 0.5



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Tests conversions to/from pyQuil circuits."""

import numpy as np
from pyquil import Program
from pyquil.gates import CNOT, CZ, RZ, H, X, Y, Z

from mitiq.interface.mitiq_pyquil.conversions import from_pyquil, to_pyquil


def test_to_pyquil_from_pyquil_simple():
    p = Program()
    p += X(0)
    p += Y(1)
    p += Z(2)
    p += CNOT(0, 1)
    p += CZ(1, 2)
    assert p.out() == to_pyquil(from_pyquil(p)).out()


def maxcut_qaoa_program(gamma: float) -> Program:
    """
    Generates a 2Q MAXCUT QAOA circuit with beta = pi/8 and with the provided
    gamma.

    Args:
        gamma: One of the two variational parameters (the other is fixed).
    Returns:
        A 2Q MAXCUT QAOA circuit with fixed beta and gamma.
    """
    q0, q1 = (0, 1)
    p = Program()
    p += H(q0)
    p += H(q1)
    p += CNOT(q0, q1)
    p += RZ(2 * gamma, q1)
    p += CNOT(q0, q1)
    p += H(q0)
    p += H(q1)
    p += RZ(np.pi / 4, q0)
    p += RZ(np.pi / 4, q1)
    p += H(q0)
    p += H(q1)
    return p


def test_to_pyquil_from_pyquil_parameterized():
    p = maxcut_qaoa_program(np.pi)
    assert p.out() == to_pyquil(from_pyquil(p)).out()


MEASURELESS_QUIL_PROGRAM = """
I 0
I 1
I 2
X 0
Y 1
Z 2
RX(1.5707963267948966) 0
RY(1.5707963267948966) 1
RZ(1.5707963267948966) 2
H 0
CZ 0 1
CNOT 1 2
CPHASE(1.5707963267948966) 0 1
ISWAP 0 1
SWAP 1 2
XY(1.5707963267948966) 0 1
CCNOT 0 1 2
CSWAP 0 1 2
"""


def test_to_pyquil_from_pyquil_almost_all_gates():
    """PHASE, PSWAP, CHPASEij, S, T, declaration, and measurement don't convert
    back and forth perfectly (in terms of labels -- the program unitaries and
    number of measurements are equivalent)."""
    p = Program(MEASURELESS_QUIL_PROGRAM)
    assert p.out() == to_pyquil(from_pyquil(p)).out()


def test_to_pyquil_from_pyquil_not_starting_at_zero():
    p = Program()
    p += X(10)
    p += Y(11)
    p += Z(12)
    p += CNOT(10, 11)
    p += CZ(11, 12)
    assert p.out() == to_pyquil(from_pyquil(p)).out()
